#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Stub version of the mail API, writes email to logs.

Can optionally send real email via SMTP or sendmail.
"""










import base64
import collections
import functools
import logging
import re
import smtplib
import subprocess

import six
from six.moves import filter

from google.appengine.api import apiproxy_stub
from google.appengine.api import mail
from google.appengine.api import mail_service_pb2
from google.appengine.runtime import apiproxy_errors


MAX_REQUEST_SIZE = 32 << 20

JAVA_TO_PYTHON_LOG_LEVELS = {
    'SEVERE': logging.CRITICAL,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'CONFIG': logging.DEBUG,
    'FINE': logging.DEBUG,
    'FINER': logging.DEBUG,
    'FINEST': logging.DEBUG,
    'ALL': logging.NOTSET,
}


class MailServiceStub(apiproxy_stub.APIProxyStub):
  """Python only mail service stub.

  Args:
    host: Host of SMTP server to use.  Blank disables sending SMTP.
    port: Port of SMTP server to use.
    user: User to log in to SMTP server as.
    password: Password for SMTP server user.
  """

  THREADSAFE = True

  def __init__(self,
               host=None,
               port=25,
               user='',
               password='',
               enable_sendmail=False,
               show_mail_body=False,
               service_name='mail',
               allow_tls=True):
    """Constructor.

    Args:
      host: Host of SMTP mail server.
      port: Port of SMTP mail server.
      user: Sending user of SMTP mail.
      password: SMTP password.
      enable_sendmail: Whether sendmail enabled or not.
      show_mail_body: Whether to show mail body in log.
      service_name: Service name expected for all calls.
      allow_tls: Allow TLS support. If `True`, use TLS provided the server
        announces support in the `EHLO` response. If `False`, do not use TLS.
    """
    super(MailServiceStub, self).__init__(service_name,
                                          max_request_size=MAX_REQUEST_SIZE)
    self._smtp_host = host
    self._smtp_port = port
    self._smtp_user = user
    self._smtp_password = password
    self._enable_sendmail = enable_sendmail
    self._show_mail_body = show_mail_body
    self._allow_tls = allow_tls

    self._cached_messages = []
    self._log_fn = logging.info






    self._saved_log_level = 'INFO'

  def _GenerateLog(self, method, message, log):
    """Generate a list of log messages representing sent mail.

    Args:
      message: Message to write to log.
      log: Log function of type string -> None
    """
    log_message = []
    log_message.append('MailService.%s' % method)
    log_message.append('  From: %s' % message.Sender)


    for address in message.To:
      log_message.append('  To: %s' % address)
    for address in message.Cc:
      log_message.append('  Cc: %s' % address)
    for address in message.Bcc:
      log_message.append('  Bcc: %s' % address)

    if message.ReplyTo:
      log_message.append('  Reply-to: %s' % message.ReplyTo)


    log_message.append('  Subject: %s' % message.Subject)


    if message.HasField('TextBody'):
      log_message.append('  Body:')
      log_message.append('    Content-type: text/plain')
      log_message.append('    Data length: %d' % len(message.TextBody))
      if self._show_mail_body:
        log_message.append('-----\n' + message.TextBody + '\n-----')


    if message.HasField('AmpHtmlBody'):
      log_message.append('  Body:')
      log_message.append('    Content-type: text/x-amp-html')
      log_message.append('    Data length: %d' % len(message.AmpHtmlBody))
      if self._show_mail_body:
        log_message.append('-----\n' + message.AmpHtmlBody + '\n-----')


    if message.HasField('HtmlBody'):
      log_message.append('  Body:')
      log_message.append('    Content-type: text/html')
      log_message.append('    Data length: %d' % len(message.HtmlBody))
      if self._show_mail_body:
        log_message.append('-----\n' + message.HtmlBody + '\n-----')


    for attachment in message.Attachment:
      log_message.append('  Attachment:')
      log_message.append('    File name: %s' % attachment.FileName)
      log_message.append('    Data length: %s' % len(attachment.Data))

    log('\n'.join(log_message))

  @apiproxy_stub.Synchronized
  def _CacheMessage(self, message):
    """Cache a message that were sent for later inspection.

    Args:
      message: A mail.MailMessage or mail.AdminMailMessage to cache.
    """
    self._cached_messages.append(message)

  def _email_message_from_mime_message(self, mime_message, mail_message_proto):
    """Construct a mail.EmailMessage from the given mime message and protobuf.

    Unlike the main constructor for `EmailMessage`, this method ensures that
    optional headers from the request protobuf are properly set on the
    `EmailMessage`.

    Args:
      mime_message: An email.MIMEMultipart instance.
      mail_message_proto: The mail_service_pb2.MailMessage instance used to
        create the mime_message.

    Returns:
      An instance of mail.EmailMessage.
    """
    headers = collections.OrderedDict()
    for header in mail_message_proto.Header:
      headers[header.name] = header.value
    if headers:
      return mail.EmailMessage(mime_message=mime_message, headers=headers)
    else:
      return mail.EmailMessage(mime_message=mime_message)


  @apiproxy_stub.Synchronized
  def get_sent_messages(self,
                        to=None,
                        sender=None,
                        subject=None,
                        body=None,
                        html=None,
                        amp_html=None):
    """Get a list of mail messages sent via the Mail API.

    Args:
      to: A regular expression that at least one recipient must match.
      sender: A regular expression that the sender must match.
      subject: A regular expression that the message subject must match.
      body: A regular expression that the text body must match.
      html: A regular expression that the HTML body must match.
      amp_html: A regular expression that the `AMP` HTML body must match.

    Returns:
      A list of matching mail.EmailMessage or mail.AdminEmailMessage objects.
    """
    messages = self._cached_messages[:]

    def recipient_matches(recipient):
      return re.search(to, recipient)

    def regex_filter(regex_pattern, message_field):
      filtered_messages = []
      for m in messages:
        message_field_contents = getattr(m, message_field, '')
        if isinstance(message_field_contents, mail.EncodedPayload):
          message_field_contents = message_field_contents.decode()
        if re.search(regex_pattern, message_field_contents):
          filtered_messages.append(m)
      return filtered_messages

    if to:
      filtered_messages = []
      for m in messages:
        to_field = getattr(m, 'to', '')
        if isinstance(to_field, six.string_types):
          to_field = [to_field]
        if list(filter(recipient_matches, to_field)):
          filtered_messages.append(m)
      messages = filtered_messages
    if sender:
      messages = regex_filter(sender, 'sender')
    if subject:
      messages = regex_filter(subject, 'subject')
    if body:
      messages = regex_filter(body, 'body')
    if html:
      messages = regex_filter(html, 'html')
    if amp_html:
      messages = regex_filter(amp_html, 'amp_html')

    return messages

  @apiproxy_stub.Synchronized
  def Clear(self):
    self._cached_messages = []

  def _SendSMTP(self, mime_message, smtp_lib=smtplib.SMTP):
    """Send MIME message via SMTP.

    Connects to SMTP server and sends MIME message.  If user is supplied
    will try to login to that server to send as authenticated.  Does not
    currently support encryption.

    Args:
      mime_message: `MimeMessage` to send.  Create using `ToMIMEMessage`.
      smtp_lib: Class of SMTP library.  Used for dependency injection.
    """
    smtp = smtp_lib()
    try:
      smtp.connect(self._smtp_host, self._smtp_port)

      smtp.ehlo_or_helo_if_needed()
      if self._allow_tls and smtp.has_extn('STARTTLS'):
        smtp.starttls()


        smtp.ehlo()
      if self._smtp_user:
        smtp.login(self._smtp_user, self._smtp_password)


      tos = [mime_message[to] for to in ['To', 'Cc', 'Bcc'] if mime_message[to]]
      smtp.sendmail(mime_message['From'], tos, mime_message.as_string())
    finally:
      smtp.quit()

  def _SendSendmail(self, mime_message,
                    popen=subprocess.Popen,
                    sendmail_command='sendmail'):
    """Send MIME message via sendmail, if exists on computer.

    Attempts to send email via sendmail.  Any `IO` failure, including
    the program not being found is ignored.

    Args:
      mime_message: `MimeMessage` to send.  Create using `ToMIMEMessage`.
      popen: popen function to create a new sub-process.
    """
    try:



      tos = []
      for to in ('To', 'Cc', 'Bcc'):
        if mime_message[to]:
          tos.extend("'%s'" % addr.strip().replace("'", r"'\''")
                     for addr in six.text_type(mime_message[to]).split(','))

      command = '%s %s' % (sendmail_command, ' '.join(tos))

      try:
        child = popen(command,
                      shell=True,
                      stdin=subprocess.PIPE,
                      stdout=subprocess.PIPE)
      except (IOError, OSError) as e:
        logging.error('Unable to open pipe to sendmail')
        raise
      try:
        child.stdin.write(
            six.ensure_binary(mime_message.as_string(), encoding='utf-8'))
        child.stdin.close()
      finally:


        while child.poll() is None:
          child.stdout.read(100)
        child.stdout.close()
    except (IOError, OSError) as e:
      logging.error('Error sending mail using sendmail: ' + str(e))

  def _Send(self, request, response, log=None,
            smtp_lib=smtplib.SMTP,
            popen=subprocess.Popen,
            sendmail_command='sendmail'):
    """Implementation of MailServer::Send().

    Logs email message.  Contents of attachments are not shown, only
    their sizes.  If SMTP is configured, will send via SMTP, else
    will use Sendmail if it is installed.

    Args:
      request: The message to send, a mail_service_pb2.MailMessage.
      response: An unused api_base_pb2.VoidProto.
      log: Log function to send log information.  Used for dependency
        injection.
      smtp_lib: Class of SMTP library.  Used for dependency injection.
      popen2: popen2 function to use for opening pipe to other process.
        Used for dependency injection.
    """
    log = log or self._log_fn

    self._ValidateExtensions(request)
    mime_message = mail.mail_message_to_mime_message(request)
    self._CacheMessage(
        self._email_message_from_mime_message(mime_message, request))
    self._GenerateLog('Send', request, log)

    if self._smtp_host and self._enable_sendmail:
      log('Both SMTP and sendmail are enabled.  Ignoring sendmail.')

    _Base64EncodeAttachments(mime_message)
    if self._smtp_host:

      self._SendSMTP(mime_message, smtp_lib)
    elif self._enable_sendmail:
      self._SendSendmail(mime_message, popen, sendmail_command)
    else:

      logging.info('You are not currently sending out real email.  '
                   'If you have sendmail installed you can use it '
                   'by using the server with --enable_sendmail')

  _Dynamic_Send = _Send

  def _SendToAdmins(self, request, response, log=None):
    """Implementation of MailServer::SendToAdmins().

    Logs email message.  Contents of attachments are not shown, only
    their sizes.

    Given the difficulty of determining who the actual sender
    is, Sendmail and SMTP are disabled for this action.

    Args:
      request: The message to send, a mail_service_pb2.MailMessage.
      response: An unused api_base_pb2.VoidProto.
      log: Log function to send log information.  Used for dependency
        injection.
    """
    log = log or self._log_fn

    self._ValidateExtensions(request)
    mime_message = mail.mail_message_to_mime_message(request)
    self._CacheMessage(mail.AdminEmailMessage(mime_message=mime_message))
    self._GenerateLog('SendToAdmins', request, log)

    if self._smtp_host and self._enable_sendmail:
      log('Both SMTP and sendmail are enabled.  Ignoring sendmail.')

  _Dynamic_SendToAdmins = _SendToAdmins

  def _Dynamic_GetSentMessages(self, request, response):
    """Get a list of all mail messages sent via the Mail API.

    Used by the Java `LocalMailService` to retrieve all sent messages.

    Args:
      request: An unused api_base_pb2.VoidProto.
      response: An instance of
        mail_stub_service_pb2.GetSentMessagesResponse.
    """
    messages = self.get_sent_messages()
    for m in messages:
      new_message = response.sent_message.add()
      new_message.MergeFrom(m.ToProto())

  def _Dynamic_ClearSentMessages(self, request, response):
    """Clear the list of sent messages and return the number cleared.

    Args:
      request: An unused api_base_pb2.VoidProto.
      response: An instance of
        mail_stub_service_pb2.ClearSentMessagesResponse.
    """
    original_messages_count = len(self._cached_messages)
    self.Clear()
    response.messages_cleared = original_messages_count

  def _Dynamic_GetLogMailBody(self, request, response):
    response.log_mail_body = self._show_mail_body

  def _Dynamic_SetLogMailBody(self, request, response):
    self._show_mail_body = request.log_mail_body

  def _Dynamic_GetLogMailLevel(self, unused_request, response):
    response.log_mail_level = self._saved_log_level

  def _Dynamic_SetLogMailLevel(self, request, unused_response):
    log_level = request.log_mail_level
    self._saved_log_level = log_level




    try:
      int_log_level = int(log_level)
      self._log_fn = functools.partial(logging.log, int_log_level)
    except ValueError:




      if log_level == 'OFF':
        def logNoOp(unused_msg, *unused_args, **unused_keywords):
          pass
        self._log_fn = logNoOp
      elif log_level in JAVA_TO_PYTHON_LOG_LEVELS:
        self._log_fn = functools.partial(logging.log,
                                         JAVA_TO_PYTHON_LOG_LEVELS[log_level])

  def _ValidateExtensions(self, request):
    """Implementation of MailServer::Send().

    Args:
      request: The message to validate, a `SendMailRequest`.

    Raises:
      mail_errors.InvalidAttachmentTypeError if the attachment type is invalid.
    """
    attachments = request.Attachment
    for attachment in attachments:
      filename = attachment.FileName
      if filename.startswith('.'):
        raise apiproxy_errors.ApplicationError(
            mail_service_pb2.MailServiceError.INVALID_ATTACHMENT_TYPE,
            'Invalid attachment type; attachments cannot be hidden files. '
            'Received \'%s\'.' % filename)
      if '.' not in filename:
        raise apiproxy_errors.ApplicationError(
            mail_service_pb2.MailServiceError.INVALID_ATTACHMENT_TYPE,
            'Invalid attachment type; attachments must have a file extension, '
            'Received \'%s\'.' % filename)
      extension = filename.rsplit('.')[-1].lower()
      if extension in mail.EXTENSION_BLACKLIST:
        raise apiproxy_errors.ApplicationError(
            mail_service_pb2.MailServiceError.INVALID_ATTACHMENT_TYPE,
            'Invalid attachment type; attachments must not have a file '
            'extension that is blacklisted. Received \'%s\'. The '
            'extension type must be one of %s.' %
            (filename, ','.join(mail.EXTENSION_BLACKLIST)))


def _Base64EncodeAttachments(mime_message):
  """Base64 encode all individual attachments that are not text.

  Args:
    mime_message: `MimeMessage` to process.
  """
  for item in mime_message.get_payload():
    if (item.get_content_maintype() not in ['multipart', 'text'] and
        'Content-Transfer-Encoding' not in item):










      original_payload = item.get_payload()
      encoded_payload = base64.b64encode(six.ensure_binary(original_payload))
      item.set_payload(encoded_payload)
      item['Content-Transfer-Encoding'] = 'base64'



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0google/appengine/api/images/images_service.proto\x12\x10google.appengine\"\xc8\x01\n\x12ImagesServiceError\"\xb1\x01\n\tErrorCode\x12\x15\n\x11UNSPECIFIED_ERROR\x10\x01\x12\x16\n\x12\x42\x41\x44_TRANSFORM_DATA\x10\x02\x12\r\n\tNOT_IMAGE\x10\x03\x12\x12\n\x0e\x42\x41\x44_IMAGE_DATA\x10\x04\x12\x13\n\x0fIMAGE_TOO_LARGE\x10\x05\x12\x14\n\x10INVALID_BLOB_KEY\x10\x06\x12\x11\n\rACCESS_DENIED\x10\x07\x12\x14\n\x10OBJECT_NOT_FOUND\x10\x08\"\x80\x01\n\x16ImagesServiceTransform\"f\n\x04Type\x12\n\n\x06RESIZE\x10\x01\x12\n\n\x06ROTATE\x10\x02\x12\x13\n\x0fHORIZONTAL_FLIP\x10\x03\x12\x11\n\rVERTICAL_FLIP\x10\x04\x12\x08\n\x04\x43ROP\x10\x05\x12\x14\n\x10IM_FEELING_LUCKY\x10\x06\"\x92\x04\n\tTransform\x12\r\n\x05width\x18\x01 \x01(\x05\x12\x0e\n\x06height\x18\x02 \x01(\x05\x12\x13\n\x0b\x63rop_to_fit\x18\x0b \x01(\x08\x12\x1a\n\rcrop_offset_x\x18\x0c \x01(\x02:\x03\x30.5\x12\x1a\n\rcrop_offset_y\x18\r \x01(\x02:\x03\x30.5\x12\x0e\n\x06rotate\x18\x03 \x01(\x05\x12\x17\n\x0fhorizontal_flip\x18\x04 \x01(\x08\x12\x15\n\rvertical_flip\x18\x05 \x01(\x08\x12\x13\n\x0b\x63rop_left_x\x18\x06 \x01(\x02\x12\x12\n\ncrop_top_y\x18\x07 \x01(\x02\x12\x17\n\x0c\x63rop_right_x\x18\x08 \x01(\x02:\x01\x31\x12\x18\n\rcrop_bottom_y\x18\t \x01(\x02:\x01\x31\x12\x12\n\nautolevels\x18\n \x01(\x08\x12\x15\n\rallow_stretch\x18\x0e \x01(\x08\x12\x1c\n\x14\x64\x65precated_width_set\x18\x65 \x01(\x08\x12\x1d\n\x15\x64\x65precated_height_set\x18\x66 \x01(\x08\x12$\n\x1c\x64\x65precated_crop_offset_x_set\x18p \x01(\x08\x12$\n\x1c\x64\x65precated_crop_offset_y_set\x18q \x01(\x08\x12#\n\x1b\x64\x65precated_crop_right_x_set\x18l \x01(\x08\x12$\n\x1c\x64\x65precated_crop_bottom_y_set\x18m \x01(\x08\"r\n\tImageData\x12\x13\n\x07\x63ontent\x18\x01 \x02(\x0c\x42\x02\x08\x01\x12\x10\n\x08\x62lob_key\x18\x02 \x01(\t\x12\r\n\x05width\x18\x03 \x01(\x05\x12\x0e\n\x06height\x18\x04 \x01(\x05\x12\x1f\n\x17\x64\x65precated_blob_key_set\x18\x66 \x01(\x08\"\xe5\x02\n\rInputSettings\x12]\n\x18\x63orrect_exif_orientation\x18\x01 \x01(\x0e\x32;.google.appengine.InputSettings.ORIENTATION_CORRECTION_TYPE\x12\x16\n\x0eparse_metadata\x18\x02 \x01(\x08\x12$\n\x1ctransparent_substitution_rgb\x18\x03 \x01(\x05\x12/\n\'deprecated_correct_exif_orientation_set\x18\x65 \x01(\x08\x12\x33\n+deprecated_transparent_substitution_rgb_set\x18g \x01(\x08\"Q\n\x1bORIENTATION_CORRECTION_TYPE\x12\x19\n\x15UNCHANGED_ORIENTATION\x10\x00\x12\x17\n\x13\x43ORRECT_ORIENTATION\x10\x01\"\x8a\x01\n\x0eOutputSettings\x12=\n\tmime_type\x18\x01 \x01(\x0e\x32*.google.appengine.OutputSettings.MIME_TYPE\x12\x0f\n\x07quality\x18\x02 \x01(\x05\"(\n\tMIME_TYPE\x12\x07\n\x03PNG\x10\x00\x12\x08\n\x04JPEG\x10\x01\x12\x08\n\x04WEBP\x10\x02\"\xd6\x01\n\x16ImagesTransformRequest\x12*\n\x05image\x18\x01 \x02(\x0b\x32\x1b.google.appengine.ImageData\x12.\n\ttransform\x18\x02 \x03(\x0b\x32\x1b.google.appengine.Transform\x12\x30\n\x06output\x18\x03 \x02(\x0b\x32 .google.appengine.OutputSettings\x12.\n\x05input\x18\x04 \x01(\x0b\x32\x1f.google.appengine.InputSettings\"^\n\x17ImagesTransformResponse\x12*\n\x05image\x18\x01 \x02(\x0b\x32\x1b.google.appengine.ImageData\x12\x17\n\x0fsource_metadata\x18\x02 \x01(\t\"\xa2\x02\n\x15\x43ompositeImageOptions\x12\x14\n\x0csource_index\x18\x01 \x02(\x05\x12\x10\n\x08x_offset\x18\x02 \x02(\x05\x12\x10\n\x08y_offset\x18\x03 \x02(\x05\x12\x0f\n\x07opacity\x18\x04 \x02(\x02\x12>\n\x06\x61nchor\x18\x05 \x02(\x0e\x32..google.appengine.CompositeImageOptions.ANCHOR\"~\n\x06\x41NCHOR\x12\x0c\n\x08TOP_LEFT\x10\x00\x12\x07\n\x03TOP\x10\x01\x12\r\n\tTOP_RIGHT\x10\x02\x12\x08\n\x04LEFT\x10\x03\x12\n\n\x06\x43\x45NTER\x10\x04\x12\t\n\x05RIGHT\x10\x05\x12\x0f\n\x0b\x42OTTOM_LEFT\x10\x06\x12\n\n\x06\x42OTTOM\x10\x07\x12\x10\n\x0c\x42OTTOM_RIGHT\x10\x08\"\x90\x01\n\x0cImagesCanvas\x12\r\n\x05width\x18\x01 \x02(\x05\x12\x0e\n\x06height\x18\x02 \x02(\x05\x12\x30\n\x06output\x18\x03 \x02(\x0b\x32 .google.appengine.OutputSettings\x12\x11\n\x05\x63olor\x18\x04 \x01(\x05:\x02-1\x12\x1c\n\x14\x64\x65precated_color_set\x18h \x01(\x08\"\xae\x01\n\x16ImagesCompositeRequest\x12*\n\x05image\x18\x01 \x03(\x0b\x32\x1b.google.appengine.ImageData\x12\x38\n\x07options\x18\x02 \x03(\x0b\x32\'.google.appengine.CompositeImageOptions\x12.\n\x06\x63\x61nvas\x18\x03 \x02(\x0b\x32\x1e.google.appengine.ImagesCanvas\"E\n\x17ImagesCompositeResponse\x12*\n\x05image\x18\x01 \x02(\x0b\x32\x1b.google.appengine.ImageData\"D\n\x16ImagesHistogramRequest\x12*\n\x05image\x18\x01 \x02(\x0b\x32\x1b.google.appengine.ImageData\";\n\x0fImagesHistogram\x12\x0b\n\x03red\x18\x01 \x03(\x05\x12\r\n\x05green\x18\x02 \x03(\x05\x12\x0c\n\x04\x62lue\x18\x03 \x03(\x05\"O\n\x17ImagesHistogramResponse\x12\x34\n\thistogram\x18\x01 \x02(\x0b\x32!.google.appengine.ImagesHistogram\"F\n\x17ImagesGetUrlBaseRequest\x12\x10\n\x08\x62lob_key\x18\x01 \x02(\t\x12\x19\n\x11\x63reate_secure_url\x18\x02 \x01(\x08\"\'\n\x18ImagesGetUrlBaseResponse\x12\x0b\n\x03url\x18\x01 \x02(\t\".\n\x1aImagesDeleteUrlBaseRequest\x12\x10\n\x08\x62lob_key\x18\x01 \x02(\t\"\x1d\n\x1bImagesDeleteUrlBaseResponseB5\n\x1f\x63om.google.appengine.api.imagesB\x0fImagesServicePb\x88\x01\x01')



_IMAGESSERVICEERROR = DESCRIPTOR.message_types_by_name['ImagesServiceError']
_IMAGESSERVICETRANSFORM = DESCRIPTOR.message_types_by_name['ImagesServiceTransform']
_TRANSFORM = DESCRIPTOR.message_types_by_name['Transform']
_IMAGEDATA = DESCRIPTOR.message_types_by_name['ImageData']
_INPUTSETTINGS = DESCRIPTOR.message_types_by_name['InputSettings']
_OUTPUTSETTINGS = DESCRIPTOR.message_types_by_name['OutputSettings']
_IMAGESTRANSFORMREQUEST = DESCRIPTOR.message_types_by_name['ImagesTransformRequest']
_IMAGESTRANSFORMRESPONSE = DESCRIPTOR.message_types_by_name['ImagesTransformResponse']
_COMPOSITEIMAGEOPTIONS = DESCRIPTOR.message_types_by_name['CompositeImageOptions']
_IMAGESCANVAS = DESCRIPTOR.message_types_by_name['ImagesCanvas']
_IMAGESCOMPOSITEREQUEST = DESCRIPTOR.message_types_by_name['ImagesCompositeRequest']
_IMAGESCOMPOSITERESPONSE = DESCRIPTOR.message_types_by_name['ImagesCompositeResponse']
_IMAGESHISTOGRAMREQUEST = DESCRIPTOR.message_types_by_name['ImagesHistogramRequest']
_IMAGESHISTOGRAM = DESCRIPTOR.message_types_by_name['ImagesHistogram']
_IMAGESHISTOGRAMRESPONSE = DESCRIPTOR.message_types_by_name['ImagesHistogramResponse']
_IMAGESGETURLBASEREQUEST = DESCRIPTOR.message_types_by_name['ImagesGetUrlBaseRequest']
_IMAGESGETURLBASERESPONSE = DESCRIPTOR.message_types_by_name['ImagesGetUrlBaseResponse']
_IMAGESDELETEURLBASEREQUEST = DESCRIPTOR.message_types_by_name['ImagesDeleteUrlBaseRequest']
_IMAGESDELETEURLBASERESPONSE = DESCRIPTOR.message_types_by_name['ImagesDeleteUrlBaseResponse']
_IMAGESSERVICEERROR_ERRORCODE = _IMAGESSERVICEERROR.enum_types_by_name['ErrorCode']
_IMAGESSERVICETRANSFORM_TYPE = _IMAGESSERVICETRANSFORM.enum_types_by_name['Type']
_INPUTSETTINGS_ORIENTATION_CORRECTION_TYPE = _INPUTSETTINGS.enum_types_by_name['ORIENTATION_CORRECTION_TYPE']
_OUTPUTSETTINGS_MIME_TYPE = _OUTPUTSETTINGS.enum_types_by_name['MIME_TYPE']
_COMPOSITEIMAGEOPTIONS_ANCHOR = _COMPOSITEIMAGEOPTIONS.enum_types_by_name['ANCHOR']
ImagesServiceError = _reflection.GeneratedProtocolMessageType('ImagesServiceError', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESSERVICEERROR,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesServiceError)

ImagesServiceTransform = _reflection.GeneratedProtocolMessageType('ImagesServiceTransform', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESSERVICETRANSFORM,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesServiceTransform)

Transform = _reflection.GeneratedProtocolMessageType('Transform', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFORM,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(Transform)

ImageData = _reflection.GeneratedProtocolMessageType('ImageData', (_message.Message,), {
  'DESCRIPTOR' : _IMAGEDATA,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImageData)

InputSettings = _reflection.GeneratedProtocolMessageType('InputSettings', (_message.Message,), {
  'DESCRIPTOR' : _INPUTSETTINGS,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(InputSettings)

OutputSettings = _reflection.GeneratedProtocolMessageType('OutputSettings', (_message.Message,), {
  'DESCRIPTOR' : _OUTPUTSETTINGS,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(OutputSettings)

ImagesTransformRequest = _reflection.GeneratedProtocolMessageType('ImagesTransformRequest', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESTRANSFORMREQUEST,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesTransformRequest)

ImagesTransformResponse = _reflection.GeneratedProtocolMessageType('ImagesTransformResponse', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESTRANSFORMRESPONSE,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesTransformResponse)

CompositeImageOptions = _reflection.GeneratedProtocolMessageType('CompositeImageOptions', (_message.Message,), {
  'DESCRIPTOR' : _COMPOSITEIMAGEOPTIONS,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(CompositeImageOptions)

ImagesCanvas = _reflection.GeneratedProtocolMessageType('ImagesCanvas', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESCANVAS,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesCanvas)

ImagesCompositeRequest = _reflection.GeneratedProtocolMessageType('ImagesCompositeRequest', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESCOMPOSITEREQUEST,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesCompositeRequest)

ImagesCompositeResponse = _reflection.GeneratedProtocolMessageType('ImagesCompositeResponse', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESCOMPOSITERESPONSE,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesCompositeResponse)

ImagesHistogramRequest = _reflection.GeneratedProtocolMessageType('ImagesHistogramRequest', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESHISTOGRAMREQUEST,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesHistogramRequest)

ImagesHistogram = _reflection.GeneratedProtocolMessageType('ImagesHistogram', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESHISTOGRAM,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesHistogram)

ImagesHistogramResponse = _reflection.GeneratedProtocolMessageType('ImagesHistogramResponse', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESHISTOGRAMRESPONSE,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesHistogramResponse)

ImagesGetUrlBaseRequest = _reflection.GeneratedProtocolMessageType('ImagesGetUrlBaseRequest', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESGETURLBASEREQUEST,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesGetUrlBaseRequest)

ImagesGetUrlBaseResponse = _reflection.GeneratedProtocolMessageType('ImagesGetUrlBaseResponse', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESGETURLBASERESPONSE,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesGetUrlBaseResponse)

ImagesDeleteUrlBaseRequest = _reflection.GeneratedProtocolMessageType('ImagesDeleteUrlBaseRequest', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESDELETEURLBASEREQUEST,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesDeleteUrlBaseRequest)

ImagesDeleteUrlBaseResponse = _reflection.GeneratedProtocolMessageType('ImagesDeleteUrlBaseResponse', (_message.Message,), {
  'DESCRIPTOR' : _IMAGESDELETEURLBASERESPONSE,
  '__module__' : 'google.appengine.api.images.images_service_pb2'

  })
_sym_db.RegisterMessage(ImagesDeleteUrlBaseResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037com.google.appengine.api.imagesB\017ImagesServicePb\210\001\001'
  _IMAGEDATA.fields_by_name['content']._options = None
  _IMAGEDATA.fields_by_name['content']._serialized_options = b'\010\001'
  _IMAGESSERVICEERROR._serialized_start=71
  _IMAGESSERVICEERROR._serialized_end=271
  _IMAGESSERVICEERROR_ERRORCODE._serialized_start=94
  _IMAGESSERVICEERROR_ERRORCODE._serialized_end=271
  _IMAGESSERVICETRANSFORM._serialized_start=274
  _IMAGESSERVICETRANSFORM._serialized_end=402
  _IMAGESSERVICETRANSFORM_TYPE._serialized_start=300
  _IMAGESSERVICETRANSFORM_TYPE._serialized_end=402
  _TRANSFORM._serialized_start=405
  _TRANSFORM._serialized_end=935
  _IMAGEDATA._serialized_start=937
  _IMAGEDATA._serialized_end=1051
  _INPUTSETTINGS._serialized_start=1054
  _INPUTSETTINGS._serialized_end=1411
  _INPUTSETTINGS_ORIENTATION_CORRECTION_TYPE._serialized_start=1330
  _INPUTSETTINGS_ORIENTATION_CORRECTION_TYPE._serialized_end=1411
  _OUTPUTSETTINGS._serialized_start=1414
  _OUTPUTSETTINGS._serialized_end=1552
  _OUTPUTSETTINGS_MIME_TYPE._serialized_start=1512
  _OUTPUTSETTINGS_MIME_TYPE._serialized_end=1552
  _IMAGESTRANSFORMREQUEST._serialized_start=1555
  _IMAGESTRANSFORMREQUEST._serialized_end=1769
  _IMAGESTRANSFORMRESPONSE._serialized_start=1771
  _IMAGESTRANSFORMRESPONSE._serialized_end=1865
  _COMPOSITEIMAGEOPTIONS._serialized_start=1868
  _COMPOSITEIMAGEOPTIONS._serialized_end=2158
  _COMPOSITEIMAGEOPTIONS_ANCHOR._serialized_start=2032
  _COMPOSITEIMAGEOPTIONS_ANCHOR._serialized_end=2158
  _IMAGESCANVAS._serialized_start=2161
  _IMAGESCANVAS._serialized_end=2305
  _IMAGESCOMPOSITEREQUEST._serialized_start=2308
  _IMAGESCOMPOSITEREQUEST._serialized_end=2482
  _IMAGESCOMPOSITERESPONSE._serialized_start=2484
  _IMAGESCOMPOSITERESPONSE._serialized_end=2553
  _IMAGESHISTOGRAMREQUEST._serialized_start=2555
  _IMAGESHISTOGRAMREQUEST._serialized_end=2623
  _IMAGESHISTOGRAM._serialized_start=2625
  _IMAGESHISTOGRAM._serialized_end=2684
  _IMAGESHISTOGRAMRESPONSE._serialized_start=2686
  _IMAGESHISTOGRAMRESPONSE._serialized_end=2765
  _IMAGESGETURLBASEREQUEST._serialized_start=2767
  _IMAGESGETURLBASEREQUEST._serialized_end=2837
  _IMAGESGETURLBASERESPONSE._serialized_start=2839
  _IMAGESGETURLBASERESPONSE._serialized_end=2878
  _IMAGESDELETEURLBASEREQUEST._serialized_start=2880
  _IMAGESDELETEURLBASEREQUEST._serialized_end=2926
  _IMAGESDELETEURLBASERESPONSE._serialized_start=2928
  _IMAGESDELETEURLBASERESPONSE._serialized_end=2957




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""CronInfo tools.

A library for working with CronInfo records, describing cron entries for an
application. Supports loading the records from yaml.
"""

import datetime
import logging
import sys
import traceback

import pytz
import six

from google.appengine.api import appinfo
from google.appengine.api import validation
from google.appengine.api import yaml_builder
from google.appengine.api import yaml_listener
from google.appengine.api import yaml_object


try:
  from google.appengine.cron import groc
  from google.appengine.cron import groctimespecification
except ImportError:
  groc = None
  groctimespecification = None









_URL_REGEX = r'^/.*$'
_TIMEZONE_REGEX = r'^.{0,100}$'
_DESCRIPTION_REGEX = r'^.{0,499}$'


SERVER_ID_RE_STRING = r'(?!-)[a-z\d\-]{1,63}'


SERVER_VERSION_RE_STRING = r'(?!-)[a-z\d\-]{1,100}'


_VERSION_REGEX = r'^(?:(?:(%s):)?)(%s)$' % (SERVER_ID_RE_STRING,
                                            SERVER_VERSION_RE_STRING)




class GrocValidator(validation.Validator):
  """Checks that a schedule is in valid groc format."""

  def Validate(self, value, key=None):
    """Validates a schedule."""
    if value is None:
      raise validation.MissingAttribute('schedule must be specified')
    if not isinstance(value, six.string_types):
      raise TypeError('schedule must be a string, not \'%r\''%type(value))
    if groc and groctimespecification:
      try:
        groctimespecification.GrocTimeSpecification(value)
      except groc.GrocException as e:
        raise validation.ValidationError('schedule \'%s\' failed to parse: %s'%(
            value, e.args[0]))
    return value


class TimezoneValidator(validation.Validator):
  """Checks that a timezone can be correctly parsed and is known."""

  def Validate(self, value, key=None):
    """Validates a timezone."""
    if not isinstance(value, six.string_types):
      raise TypeError('timezone must be a string, not \'%r\'' % type(value))
    if pytz is None:

      return value
    try:
      pytz.timezone(value)
    except pytz.UnknownTimeZoneError:
      raise validation.ValidationError('timezone \'%s\' is unknown' % value)
    except IOError:


      return value
    except:


      unused_e, v, t = sys.exc_info()
      logging.warning('pytz raised an unexpected error: %s.\n' % (v) +
                      'Traceback:\n' + '\n'.join(traceback.format_tb(t)))
      raise
    return value


CRON = 'cron'

URL = 'url'
SCHEDULE = 'schedule'
TIMEZONE = 'timezone'
DESCRIPTION = 'description'
TARGET = 'target'

RETRY_PARAMETERS = 'retry_parameters'
JOB_RETRY_LIMIT = 'job_retry_limit'
JOB_AGE_LIMIT = 'job_age_limit'
MIN_BACKOFF_SECONDS = 'min_backoff_seconds'
MAX_BACKOFF_SECONDS = 'max_backoff_seconds'
MAX_DOUBLINGS = 'max_doublings'

ATTEMPT_DEADLINE = 'attempt_deadline'


class AttemptDeadlineValidator(validation.Validator):
  """Checks that attempt deadline can be correctly parsed."""

  def Validate(self, value, key='attempt_deadline'):
    """Validates that the input is of the given format and returns the results as timedelta."""

    value = value.strip()

    if value.endswith('s'):
      try:
        return datetime.timedelta(seconds=float(value[:-1]))
      except Exception as e:
        raise validation.ValidationError(
            'Value %s cannot be assigned to attempt_deadline: %s' % (value, e))
    raise validation.ValidationError(
        'Value %s cannot be assigned to attempt_deadline. String should be formatted as a floating point number followed by \'s\'.'
        % value)


class MalformedCronfigurationFile(Exception):
  """Configuration file for Cron is malformed."""
  pass


class RetryParameters(validation.Validated):
  """Retry parameters for a single cron job."""
  ATTRIBUTES = {
      JOB_RETRY_LIMIT: validation.Optional(
          validation.Range(minimum=0,

                           maximum=sys.maxsize,
                           range_type=int)),
      JOB_AGE_LIMIT: validation.Optional(validation.TimeValue()),
      MIN_BACKOFF_SECONDS: validation.Optional(
          validation.Range(0.0, None, range_type=float)),
      MAX_BACKOFF_SECONDS: validation.Optional(
          validation.Range(0.0, None, range_type=float)),
      MAX_DOUBLINGS: validation.Optional(
          validation.Range(0, None, range_type=int)),
  }


class CronEntry(validation.Validated):
  """A cron entry describes a single cron job."""
  ATTRIBUTES = {
      URL: _URL_REGEX,
      SCHEDULE: GrocValidator(),
      TIMEZONE: validation.Optional(TimezoneValidator()),
      DESCRIPTION: validation.Optional(_DESCRIPTION_REGEX),
      RETRY_PARAMETERS: validation.Optional(RetryParameters),
      TARGET: validation.Optional(_VERSION_REGEX),
      ATTEMPT_DEADLINE: validation.Optional(AttemptDeadlineValidator())
  }


class CronInfoExternal(validation.Validated):
  """CronInfoExternal describes all cron entries for an application."""
  ATTRIBUTES = {
      appinfo.APPLICATION: validation.Optional(appinfo.APPLICATION_RE_STRING),
      CRON: validation.Optional(validation.Repeated(CronEntry))
  }


def LoadSingleCron(cron_info, open_fn=None):
  """Load a cron.yaml file or string and return a CronInfoExternal object."""
  del open_fn
  builder = yaml_object.ObjectBuilder(CronInfoExternal)
  handler = yaml_builder.BuilderHandler(builder)
  listener = yaml_listener.EventListener(handler)
  listener.Parse(cron_info)

  cron_info_result = handler.GetResults()
  if len(cron_info_result) < 1:
    raise MalformedCronfigurationFile('Empty cron configuration.')
  if len(cron_info_result) > 1:
    raise MalformedCronfigurationFile('Multiple cron sections '
                                      'in configuration.')
  return cron_info_result[0]



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#














"""Context class."""

import logging
import sys

from google.appengine.ext.ndb import eventloop
from google.appengine.ext.ndb import key as key_module
from google.appengine.ext.ndb import model
from google.appengine.ext.ndb import tasklets
from google.appengine.ext.ndb import utils
import six
from six.moves import range
from six.moves import zip

from google.appengine.api import datastore
from google.appengine.api import datastore_errors
from google.appengine.api import memcache
from google.appengine.api import namespace_manager
from google.appengine.api import urlfetch
from google.appengine.datastore import datastore_rpc
from google.protobuf import message
from google.appengine.datastore import entity_bytes_pb2 as entity_pb2

__all__ = ['Context', 'ContextOptions', 'TransactionOptions', 'AutoBatcher',
           'EVENTUAL_CONSISTENCY',
          ]

_LOCK_TIME = 32
_LOCKED = 0



EVENTUAL_CONSISTENCY = datastore_rpc.Configuration.EVENTUAL_CONSISTENCY


class ContextOptions(datastore_rpc.Configuration):
  """Configuration options that may be passed along with get/put/delete."""

  @datastore_rpc.ConfigOption
  def use_cache(value):
    if not isinstance(value, bool):
      raise datastore_errors.BadArgumentError(
          'use_cache should be a bool (%r)' % (value,))
    return value

  @datastore_rpc.ConfigOption
  def use_memcache(value):
    if not isinstance(value, bool):
      raise datastore_errors.BadArgumentError(
          'use_memcache should be a bool (%r)' % (value,))
    return value

  @datastore_rpc.ConfigOption
  def use_datastore(value):
    if not isinstance(value, bool):
      raise datastore_errors.BadArgumentError(
          'use_datastore should be a bool (%r)' % (value,))
    return value

  @datastore_rpc.ConfigOption
  def memcache_timeout(value):
    if not isinstance(value, six.integer_types):
      raise datastore_errors.BadArgumentError(
          'memcache_timeout should be an integer (%r)' % (value,))
    return value

  @datastore_rpc.ConfigOption
  def max_memcache_items(value):
    if not isinstance(value, six.integer_types):
      raise datastore_errors.BadArgumentError(
          'max_memcache_items should be an integer (%r)' % (value,))
    return value

  @datastore_rpc.ConfigOption
  def memcache_deadline(value):
    if not isinstance(value, six.integer_types):
      raise datastore_errors.BadArgumentError(
          'memcache_deadline should be an integer (%r)' % (value,))
    return value


class TransactionOptions(ContextOptions, datastore_rpc.TransactionOptions):
  """Support both context options and transaction options."""



_OPTION_TRANSLATIONS = {
    'options': 'config',
}


def _make_ctx_options(ctx_options, config_cls=ContextOptions):
  """Helper to construct a ContextOptions object from keyword arguments.

  Args:
    ctx_options: A dict of keyword arguments.
    config_cls: Optional Configuration class to use, default ContextOptions.

  Note that either 'options' or 'config' can be used to pass another
  Configuration object, but not both.  If another Configuration
  object is given it provides default values.

  Returns:
    A Configuration object, or None if ctx_options is empty.
  """
  if not ctx_options:
    return None
  for key in list(ctx_options):
    translation = _OPTION_TRANSLATIONS.get(key)
    if translation:
      if translation in ctx_options:
        raise ValueError('Cannot specify %s and %s at the same time' %
                         (key, translation))
      ctx_options[translation] = ctx_options.pop(key)
  return config_cls(**ctx_options)


class AutoBatcher(object):
  """Batches multiple async calls if they share the same RPC options.

  Here is an example to explain what this class does.

  Life of a `key.get_async(options)` API call:

  * `Key` gets the singleton `Context` instance and invokes `Context.get`.

  * `Context.get` calls `Context._get_batcher.add(key, options)`. This
     returns a future `fut` as the return value of `key.get_async`.
     At this moment, `key.get_async` returns.

  * When more than "limit" number of `_get_batcher.add()` was called,
     `_get_batcher` invokes its `self._todo_tasklet`, `Context._get_tasklet`,
     with the list of keys seen so far.

  * `Context._get_tasklet` fires a MultiRPC and waits on it.

  * Upon MultiRPC completion, `Context._get_tasklet` passes on the results
     to the respective `fut` from `key.get_async`.

  * If user calls `fut.get_result()` before "limit" number of `add()` was
     called, `fut.get_result()` will repeatedly call `eventloop.run1()`.

  * After processing immediate callbacks, `eventloop` will run idlers.
     `AutoBatcher._on_idle` is an idler.

  * `_on_idle` will run the `todo_tasklet` before the batch is full.

  So the engine is `todo_tasklet`, which is a proxy tasklet that can combine
  arguments into batches and passes along results back to respective futures.
  This class is mainly a helper that invokes `todo_tasklet` with the right
  arguments at the right time.
  """

  def __init__(self, todo_tasklet, limit):
    """Init.

    Args:
      todo_tasklet: The tasklet that actually fires RPC and waits on a MultiRPC.
        It should take a list of (future, arg) pairs and an "options" as
        arguments. "options" are rpc options.
      limit: Max number of items to batch for each distinct value of "options".
    """
    self._todo_tasklet = todo_tasklet
    self._limit = limit


    self._queues = {}
    self._running = []
    self._cache = {}

  def __repr__(self):
    return '%s(%s)' % (self.__class__.__name__, self._todo_tasklet.__name__)

  def run_queue(self, options, todo):
    """Actually run the `_todo_tasklet`."""
    utils.logging_debug('AutoBatcher(%s): %d items',
                        self._todo_tasklet.__name__, len(todo))
    batch_fut = self._todo_tasklet(todo, options)
    self._running.append(batch_fut)

    batch_fut.add_callback(self._finished_callback, batch_fut, todo)

  def _on_idle(self):
    """An idler eventloop can run.

    Eventloop calls this when it has finished processing all immediate
    callbacks. This method runs _todo_tasklet even before the batch is full.
    """
    if not self.action():
      return None
    return True


  def add(self, arg, options=None):
    """Returns back an instance of future after adding an arg.

    Args:
      arg: One argument for `_todo_tasklet`.
      options: RPC options.

    Return:
      An instance of future, representing the result of running
        `_todo_tasklet` without batching.
    """
    fut = tasklets.Future('%s.add(%s, %s)' % (self, arg, options))
    todo = self._queues.get(options)
    if todo is None:
      utils.logging_debug('AutoBatcher(%s): creating new queue for %r',
                          self._todo_tasklet.__name__, options)
      if not self._queues:
        eventloop.add_idle(self._on_idle)
      todo = self._queues[options] = []
    todo.append((fut, arg))
    if len(todo) >= self._limit:
      del self._queues[options]
      self.run_queue(options, todo)
    return fut

  def add_once(self, arg, options=None):
    cache_key = (arg, options)
    fut = self._cache.get(cache_key)
    if fut is None:
      fut = self.add(arg, options)
      self._cache[cache_key] = fut
      fut.add_immediate_callback(self._cache.__delitem__, cache_key)
    return fut

  def action(self):
    queues = self._queues
    if not queues:
      return False
    options, todo = queues.popitem()
    self.run_queue(options, todo)
    return True

  def _finished_callback(self, batch_fut, todo):
    """Passes exception along.

    Args:
      batch_fut: the batch future returned by running todo_tasklet.
      todo: (fut, option) pair. fut is the future return by each add() call.

    If the batch fut was successful, it has already called fut.set_result()
    on other individual futs. This method only handles when the batch fut
    encountered an exception.
    """
    self._running.remove(batch_fut)
    err = batch_fut.get_exception()
    if err is not None:
      tb = batch_fut.get_traceback()
      for (fut, _) in todo:
        if not fut.done():
          fut.set_exception(err, tb)

  @tasklets.tasklet
  def flush(self):
    while self._running or self.action():
      if self._running:
        yield self._running


class Context(object):

  def __init__(self, conn=None, auto_batcher_class=AutoBatcher, config=None,
               parent_context=None):


    if conn is None:
      conn = model.make_connection(config)
    self._conn = conn
    self._auto_batcher_class = auto_batcher_class
    self._parent_context = parent_context



    max_get = (datastore_rpc.Configuration.max_get_keys(config, conn.config) or
               datastore_rpc.Connection.MAX_GET_KEYS)
    max_put = (datastore_rpc.Configuration.max_put_entities(config,
                                                            conn.config) or
               datastore_rpc.Connection.MAX_PUT_ENTITIES)
    max_delete = (datastore_rpc.Configuration.max_delete_keys(config,
                                                              conn.config) or
                  datastore_rpc.Connection.MAX_DELETE_KEYS)

    self._get_batcher = auto_batcher_class(self._get_tasklet, max_get)
    self._put_batcher = auto_batcher_class(self._put_tasklet, max_put)
    self._delete_batcher = auto_batcher_class(self._delete_tasklet, max_delete)

    max_memcache = (ContextOptions.max_memcache_items(config, conn.config) or
                    datastore_rpc.Connection.MAX_GET_KEYS)

    self._memcache_get_batcher = auto_batcher_class(self._memcache_get_tasklet,
                                                    max_memcache)
    self._memcache_set_batcher = auto_batcher_class(self._memcache_set_tasklet,
                                                    max_memcache)
    self._memcache_del_batcher = auto_batcher_class(self._memcache_del_tasklet,
                                                    max_memcache)
    self._memcache_off_batcher = auto_batcher_class(self._memcache_off_tasklet,
                                                    max_memcache)

    self._batchers = [self._get_batcher,
                      self._put_batcher,
                      self._delete_batcher,
                      self._memcache_get_batcher,
                      self._memcache_set_batcher,
                      self._memcache_del_batcher,
                      self._memcache_off_batcher,
                     ]
    self._cache = {}
    self._memcache = memcache.Client()
    self._on_commit_queue = []



  _memcache_prefix = b'NDB9:'

  @tasklets.tasklet
  def flush(self):

    more = True
    while more:
      yield [batcher.flush() for batcher in self._batchers]
      more = False
      for batcher in self._batchers:
        if batcher._running or batcher._queues:
          more = True
          break

  @tasklets.tasklet
  def _get_tasklet(self, todo, options):
    if not todo:
      raise RuntimeError('Nothing to do.')

    datastore_keys = []
    for unused_fut, key in todo:
      datastore_keys.append(key)

    entities = yield self._conn.async_get(options, datastore_keys)
    for ent, (fut, unused_key) in zip(entities, todo):
      fut.set_result(ent)

  @tasklets.tasklet
  def _put_tasklet(self, todo, options):
    if not todo:
      raise RuntimeError('Nothing to do.')


    datastore_entities = []
    for unused_fut, ent in todo:
      datastore_entities.append(ent)

    keys = yield self._conn.async_put(options, datastore_entities)
    for key, (fut, ent) in zip(keys, todo):
      if key != ent._key:
        if ent._has_complete_key():
          ent_key = ent._key
          raise datastore_errors.BadKeyError(
              'Entity Key differs from the one returned by Datastore. '
              'Returned Key: %r, Entity Key: %r' % (key, ent_key))
        ent._key = key
      fut.set_result(key)

  @tasklets.tasklet
  def _delete_tasklet(self, todo, options):
    if not todo:
      raise RuntimeError('Nothing to do.')
    futures = []
    datastore_keys = []
    for fut, key in todo:
      futures.append(fut)
      datastore_keys.append(key)

    yield self._conn.async_delete(options, datastore_keys)

    for fut in futures:
      fut.set_result(None)












  @staticmethod
  def default_cache_policy(key):
    """Default cache policy.

    This defers to `_use_cache` on the `Model` class.

    Args:
      key: Key instance.

    Returns:
      A bool or `None`.
    """
    flag = None
    if key is not None:
      modelclass = model.Model._kind_map.get(key.kind())
      if modelclass is not None:
        policy = getattr(modelclass, '_use_cache', None)
        if policy is not None:
          if isinstance(policy, bool):
            flag = policy
          else:
            flag = policy(key)
    return flag

  _cache_policy = default_cache_policy

  def get_cache_policy(self):
    """Return the current context cache policy function.

    Returns:
      A function that accepts a `Key` instance as argument and returns
      a bool indicating if it should be cached.  May be `None`.
    """
    return self._cache_policy

  def set_cache_policy(self, func):
    """Set the context cache policy function.

    Args:
      func: A function that accepts a `Key` instance as argument and returns
        a bool indicating if it should be cached. May be `None`.
    """
    if func is None:
      func = self.default_cache_policy
    elif isinstance(func, bool):
      func = lambda unused_key, flag=func: flag
    self._cache_policy = func

  def _use_cache(self, key, options=None):
    """Return whether to use the context cache for this key.

    Args:
      key: Key instance.
      options: ContextOptions instance, or None.

    Returns:
      True if the key should be cached, False otherwise.
    """
    flag = ContextOptions.use_cache(options)
    if flag is None:
      flag = self._cache_policy(key)
    if flag is None:
      flag = ContextOptions.use_cache(self._conn.config)
    if flag is None:
      flag = True
    return flag

  @staticmethod
  def default_memcache_policy(key):
    """Default Memcache policy.

    This defers to `_use_memcache` on the `Model` class.

    Args:
      key: Key instance.

    Returns:
      A bool or `None`.
    """
    flag = None
    if key is not None:
      modelclass = model.Model._kind_map.get(key.kind())
      if modelclass is not None:
        policy = getattr(modelclass, '_use_memcache', None)
        if policy is not None:
          if isinstance(policy, bool):
            flag = policy
          else:
            flag = policy(key)
    return flag

  _memcache_policy = default_memcache_policy

  def get_memcache_policy(self):
    """Return the current memcache policy function.

    Returns:
      A function that accepts a `Key` instance as argument and returns
      a bool indicating if it should be cached. May be `None`.
    """
    return self._memcache_policy

  def set_memcache_policy(self, func):
    """Set the memcache policy function.

    Args:
      func: A function that accepts a Key instance as argument and returns
        a bool indicating if it should be cached.  May be None.
    """
    if func is None:
      func = self.default_memcache_policy
    elif isinstance(func, bool):
      func = lambda unused_key, flag=func: flag
    self._memcache_policy = func

  def _use_memcache(self, key, options=None):
    """Return whether to use memcache for this key.

    Args:
      key: Key instance.
      options: ContextOptions instance, or None.

    Returns:
      True if the key should be cached in memcache, False otherwise.
    """
    flag = ContextOptions.use_memcache(options)
    if flag is None:
      flag = self._memcache_policy(key)
    if flag is None:
      flag = ContextOptions.use_memcache(self._conn.config)
    if flag is None:
      flag = True
    return flag

  @staticmethod
  def default_datastore_policy(key):
    """Default Datastore policy.

    This defers to `_use_datastore` on the `Model` class.

    Args:
      key: Key instance.

    Returns:
      A bool or `None`.
    """
    flag = None
    if key is not None:
      modelclass = model.Model._kind_map.get(key.kind())
      if modelclass is not None:
        policy = getattr(modelclass, '_use_datastore', None)
        if policy is not None:
          if isinstance(policy, bool):
            flag = policy
          else:
            flag = policy(key)
    return flag

  _datastore_policy = default_datastore_policy

  def get_datastore_policy(self):
    """Return the current context datastore policy function.

    Returns:
      A function that accepts a `Key` instance as argument and returns
      a bool indicating if it should use the datastore. May be `None`.
    """
    return self._datastore_policy

  def set_datastore_policy(self, func):
    """Set the context datastore policy function.

    Args:
      func: A function that accepts a `Key` instance as argument and returns
        a bool indicating if it should use the datastore. May be `None`.
    """
    if func is None:
      func = self.default_datastore_policy
    elif isinstance(func, bool):
      func = lambda unused_key, flag=func: flag
    self._datastore_policy = func

  def _use_datastore(self, key, options=None):
    """Return whether to use the datastore for this key.

    Args:
      key: Key instance.
      options: ContextOptions instance, or None.

    Returns:
      True if the datastore should be used, False otherwise.
    """
    flag = ContextOptions.use_datastore(options)
    if flag is None:
      flag = self._datastore_policy(key)
    if flag is None:
      flag = ContextOptions.use_datastore(self._conn.config)
    if flag is None:
      flag = True
    return flag

  @staticmethod
  def default_memcache_timeout_policy(key):
    """Default Memcache timeout policy.

    This defers to `_memcache_timeout` on the `Model` class.

    Args:
      key: Key instance.

    Returns:
      Memcache timeout to use (integer), or `None`.
    """
    timeout = None
    if key is not None and isinstance(key, model.Key):
      modelclass = model.Model._kind_map.get(key.kind())
      if modelclass is not None:
        policy = getattr(modelclass, '_memcache_timeout', None)
        if policy is not None:
          if isinstance(policy, six.integer_types):
            timeout = policy
          else:
            timeout = policy(key)
    return timeout

  _memcache_timeout_policy = default_memcache_timeout_policy

  def set_memcache_timeout_policy(self, func):
    """Set the policy function for memcache timeout (expiration).

    If the function returns `0`, it implies the default timeout.

    Args:
      func: A function that accepts a key instance as argument and returns
        an integer indicating the desired memcache timeout. May be `None`.
    """
    if func is None:
      func = self.default_memcache_timeout_policy
    elif isinstance(func, six.integer_types):
      func = lambda unused_key, flag=func: flag
    self._memcache_timeout_policy = func

  def get_memcache_timeout_policy(self):
    """Return the current policy function for memcache timeout (expiration)."""
    return self._memcache_timeout_policy

  def _get_memcache_timeout(self, key, options=None):
    """Return the memcache timeout (expiration) for this key."""
    timeout = ContextOptions.memcache_timeout(options)
    if timeout is None:
      timeout = self._memcache_timeout_policy(key)
    if timeout is None:
      timeout = ContextOptions.memcache_timeout(self._conn.config)
    if timeout is None:
      timeout = 0
    return timeout

  def _get_memcache_deadline(self, options=None):
    """Return the memcache RPC deadline.

    Not to be confused with the memcache timeout, or expiration.

    This is only used by datastore operations when using memcache
    as a cache; it is ignored by the direct memcache calls.

    There is no way to vary this per key or per entity; you must either
    set it on a specific call (e.g. key.get(memcache_deadline=1) or
    in the configuration options of the context's connection.
    """

    return ContextOptions.memcache_deadline(options, self._conn.config)

  def _load_from_cache_if_available(self, key):
    """Returns a cached Model instance given the entity key if available.

    Args:
      key: Key instance.

    Returns:
      A Model instance if the key exists in the cache.
    """
    if key in self._cache:
      entity = self._cache[key]
      if entity is None or entity._key == key:


        raise tasklets.Return(entity)









  @tasklets.tasklet
  def get(self, key, **ctx_options):
    """Returns a `Model` instance given the entity key.

    It will use the context cache if the cache policy for the given
    key is enabled.

    Args:
      key: Key instance.
      **ctx_options: Context options.

    Returns:
      A `Model` instance if the key exists in the datastore; `None` otherwise.
    """
    options = _make_ctx_options(ctx_options)
    use_cache = self._use_cache(key, options)
    if use_cache:
      self._load_from_cache_if_available(key)

    use_datastore = self._use_datastore(key, options)
    if (use_datastore and
        isinstance(self._conn, datastore_rpc.TransactionalConnection)):
      use_memcache = False
    else:
      use_memcache = self._use_memcache(key, options)
    ns = key.namespace()
    memcache_deadline = None

    if use_memcache:
      mkey = self._memcache_prefix + key.urlsafe()
      memcache_deadline = self._get_memcache_deadline(options)
      mvalue = yield self.memcache_get(mkey, for_cas=use_datastore,
                                       namespace=ns, use_cache=True,
                                       deadline=memcache_deadline)

      if use_cache:
        self._load_from_cache_if_available(key)
      if mvalue not in (_LOCKED, None):
        cls = model.Model._lookup_model(key.kind(),
                                        self._conn.adapter.default_model)
        pb = entity_pb2.EntityProto()

        try:
          pb.MergeFromString(mvalue)
        except message.DecodeError:
          logging.warning('Corrupt memcache entry found '
                          'with key %s and namespace %s', mkey, ns)
          mvalue = None
        else:
          entity = cls._from_pb(pb)

          entity._key = key
          if use_cache:

            self._cache[key] = entity
          raise tasklets.Return(entity)

      if mvalue is None and use_datastore:
        yield self.memcache_set(mkey, _LOCKED, time=_LOCK_TIME, namespace=ns,
                                use_cache=True, deadline=memcache_deadline)
        yield self.memcache_gets(mkey, namespace=ns, use_cache=True,
                                 deadline=memcache_deadline)

    if not use_datastore:


      raise tasklets.Return(None)

    if use_cache:
      entity = yield self._get_batcher.add_once(key, options)
    else:
      entity = yield self._get_batcher.add(key, options)

    if entity is not None:
      if use_memcache and mvalue != _LOCKED:

        pbs = entity._to_pb(set_key=False).SerializePartialToString()





        if len(pbs) <= memcache.MAX_VALUE_SIZE:
          timeout = self._get_memcache_timeout(key, options)




          yield self.memcache_cas(mkey, pbs, time=timeout, namespace=ns,
                                  deadline=memcache_deadline)

    if use_cache:


      self._cache[key] = entity

    raise tasklets.Return(entity)

  @tasklets.tasklet
  def put(self, entity, **ctx_options):
    options = _make_ctx_options(ctx_options)


    key = entity._key
    if key is None:

      key = model.Key(entity.__class__, None)
    use_datastore = self._use_datastore(key, options)
    use_memcache = None
    memcache_deadline = None

    if entity._has_complete_key():
      use_memcache = self._use_memcache(key, options)
      if use_memcache:

        memcache_deadline = self._get_memcache_deadline(options)
        mkey = self._memcache_prefix + key.urlsafe()
        ns = key.namespace()
        if use_datastore:
          yield self.memcache_set(mkey, _LOCKED, time=_LOCK_TIME,
                                  namespace=ns, use_cache=True,
                                  deadline=memcache_deadline)
        else:
          pbs = entity._to_pb(set_key=False).SerializePartialToString()


          if len(pbs) > memcache.MAX_VALUE_SIZE:
            raise ValueError('Values may not be more than %d bytes in length; '
                             'received %d bytes' % (memcache.MAX_VALUE_SIZE,
                                                    len(pbs)))
          timeout = self._get_memcache_timeout(key, options)
          yield self.memcache_set(mkey, pbs, time=timeout, namespace=ns,
                                  deadline=memcache_deadline)

    if use_datastore:
      key = yield self._put_batcher.add(entity, options)
      if not isinstance(self._conn, datastore_rpc.TransactionalConnection):
        if use_memcache is None:
          use_memcache = self._use_memcache(key, options)
        if use_memcache:
          mkey = self._memcache_prefix + key.urlsafe()
          ns = key.namespace()

          yield self.memcache_delete(mkey, namespace=ns,
                                     deadline=memcache_deadline)

    if key is not None:
      if entity._key != key:
        logging.info('replacing key %s with %s', entity._key, key)
        entity._key = key

      if self._use_cache(key, options):

        self._cache[key] = entity

    raise tasklets.Return(key)

  @tasklets.tasklet
  def delete(self, key, **ctx_options):
    options = _make_ctx_options(ctx_options)
    if self._use_memcache(key, options):
      memcache_deadline = self._get_memcache_deadline(options)
      mkey = self._memcache_prefix + key.urlsafe()
      ns = key.namespace()

      yield self.memcache_set(mkey, _LOCKED, time=_LOCK_TIME, namespace=ns,
                              use_cache=True, deadline=memcache_deadline)

    if self._use_datastore(key, options):
      yield self._delete_batcher.add(key, options)


    if self._use_cache(key, options):
      self._cache[key] = None

  @tasklets.tasklet
  def allocate_ids(self, key, size=None, max=None, **ctx_options):
    options = _make_ctx_options(ctx_options)
    lo_hi = yield self._conn.async_allocate_ids(options, key, size, max)
    raise tasklets.Return(lo_hi)

  @tasklets.tasklet
  def get_indexes(self, **ctx_options):
    options = _make_ctx_options(ctx_options)
    index_list = yield self._conn.async_get_indexes(options)
    raise tasklets.Return(index_list)

  @utils.positional(3)
  def map_query(self, query, callback, pass_batch_into_callback=None,
                options=None, merge_future=None):
    mfut = merge_future
    if mfut is None:
      mfut = tasklets.MultiFuture('map_query')

    @tasklets.tasklet
    def helper():
      try:
        inq = tasklets.SerialQueueFuture()
        query.run_to_queue(inq, self._conn, options)
        while True:
          try:
            batch, i, ent = yield inq.getq()
          except EOFError:
            break
          ent = self._update_cache_from_query_result(ent, options)
          if ent is None:
            continue
          if callback is None:
            val = ent
          else:

            if pass_batch_into_callback:
              val = callback(batch, i, ent)
            else:
              val = callback(ent)
          mfut.putq(val)
      except GeneratorExit:
        raise
      except Exception as err:
        _, _, tb = sys.exc_info()
        mfut.set_exception(err, tb)
        raise
      else:
        mfut.complete()

    helper()
    return mfut

  def _update_cache_from_query_result(self, ent, options):
    if isinstance(ent, model.Key):
      return ent
    if ent._projection:
      return ent
    key = ent._key
    if not self._use_cache(key, options):
      return ent



    if key in self._cache:
      cached_ent = self._cache[key]
      if (cached_ent is None or
          cached_ent.key == key and cached_ent.__class__ is ent.__class__):
        return cached_ent


    self._cache[key] = ent
    return ent

  @utils.positional(2)
  def iter_query(self, query, callback=None, pass_batch_into_callback=None,
                 options=None):
    return self.map_query(query, callback=callback, options=options,
                          pass_batch_into_callback=pass_batch_into_callback,
                          merge_future=tasklets.SerialQueueFuture())

  @tasklets.tasklet
  def transaction(self, callback, **ctx_options):



    options = _make_ctx_options(ctx_options, TransactionOptions)
    propagation = TransactionOptions.propagation(options)
    if propagation is None:
      propagation = TransactionOptions.NESTED

    mode = datastore_rpc.TransactionMode.READ_WRITE
    if ctx_options.get('read_only', False):
      mode = datastore_rpc.TransactionMode.READ_ONLY

    parent = self
    if propagation == TransactionOptions.NESTED:
      if self.in_transaction():
        raise datastore_errors.BadRequestError(
            'Nested transactions are not supported.')
    elif propagation == TransactionOptions.MANDATORY:
      if not self.in_transaction():
        raise datastore_errors.BadRequestError(
            'Requires an existing transaction.')
      result = callback()
      if isinstance(result, tasklets.Future):
        result = yield result
      raise tasklets.Return(result)
    elif propagation == TransactionOptions.ALLOWED:
      if self.in_transaction():
        result = callback()
        if isinstance(result, tasklets.Future):
          result = yield result
        raise tasklets.Return(result)
    elif propagation == TransactionOptions.INDEPENDENT:
      while parent.in_transaction():
        parent = parent._parent_context
        if parent is None:
          raise datastore_errors.BadRequestError(
              'Context without non-transactional ancestor')
    else:
      raise datastore_errors.BadArgumentError(
          'Invalid propagation value (%s).' % (propagation,))

    app = TransactionOptions.app(options) or key_module._DefaultAppId()

    retries = TransactionOptions.retries(options)
    if retries is None:
      retries = 3
    yield parent.flush()

    transaction = None
    tconn = None
    for _ in range(1 + max(0, retries)):
      previous_transaction = (
          transaction
          if mode == datastore_rpc.TransactionMode.READ_WRITE else None)
      transaction = yield (parent._conn.async_begin_transaction(
          options, app,
          previous_transaction,
          mode))
      tconn = datastore_rpc.TransactionalConnection(
          adapter=parent._conn.adapter,
          config=parent._conn.config,
          transaction=transaction,
          _api_version=parent._conn._api_version)
      tctx = parent.__class__(conn=tconn,
                              auto_batcher_class=parent._auto_batcher_class,
                              parent_context=parent)
      tctx._old_ds_conn = datastore._GetConnection()
      ok = False
      try:





        tctx.set_memcache_policy(parent.get_memcache_policy())
        tctx.set_memcache_timeout_policy(parent.get_memcache_timeout_policy())
        tasklets.set_context(tctx)
        datastore._SetConnection(tconn)
        try:
          try:
            result = callback()
            if isinstance(result, tasklets.Future):
              result = yield result
          finally:
            yield tctx.flush()
        except GeneratorExit:
          raise
        except Exception:
          t, e, tb = sys.exc_info()
          tconn.async_rollback(options)
          if issubclass(t, datastore_errors.Rollback):

            return
          else:
            six.reraise(t, e, tb)
        else:
          ok = yield tconn.async_commit(options)
          if ok:
            parent._cache.update(tctx._cache)
            yield parent._clear_memcache(tctx._cache)
            raise tasklets.Return(result)

      finally:
        datastore._SetConnection(tctx._old_ds_conn)
        del tctx._old_ds_conn
        if ok:





          for on_commit_callback in tctx._on_commit_queue:
            on_commit_callback()


    tconn.async_rollback(options)
    raise datastore_errors.TransactionFailedError(
        'The transaction could not be committed. Please try again.')

  def in_transaction(self):
    """Return whether a transaction is currently active."""
    return isinstance(self._conn, datastore_rpc.TransactionalConnection)

  def call_on_commit(self, callback):
    """Call a callback upon successful commit of a transaction.

    If not in a transaction, the callback is called immediately.

    In a transaction, multiple callbacks may be registered and will be
    called once the transaction commits, in the order in which they
    were registered.  If the transaction fails, the callbacks will not
    be called.

    If the callback raises an exception, it bubbles up normally.  This
    means:

    - If the callback is called immediately, any exception it
    raises will bubble up immediately.
    - If the call is postponed until
    commit, remaining callbacks will be skipped and the exception will
    bubble up through the `transaction()` call. However, the
    transaction is already committed at that point.
    """
    if not self.in_transaction():
      callback()
    else:
      self._on_commit_queue.append(callback)

  def clear_cache(self):
    """Clears the in-memory cache.

    NOTE: This does not affect memcache.
    """
    self._cache.clear()

  @tasklets.tasklet
  def _clear_memcache(self, keys):
    keys = set(key for key in keys if self._use_memcache(key))
    futures = []
    for key in keys:
      mkey = self._memcache_prefix + key.urlsafe()
      ns = key.namespace()
      fut = self.memcache_delete(mkey, namespace=ns)
      futures.append(fut)
    yield futures

  @tasklets.tasklet
  def _memcache_get_tasklet(self, todo, options):
    if not todo:
      raise RuntimeError('Nothing to do.')
    for_cas, namespace, deadline = options
    keys = set()
    for unused_fut, key in todo:
      keys.add(key)
    rpc = memcache.create_rpc(deadline=deadline)
    results = yield self._memcache.get_multi_async(keys, for_cas=for_cas,
                                                   namespace=namespace,
                                                   rpc=rpc)
    for fut, key in todo:
      fut.set_result(results.get(key))

  @tasklets.tasklet
  def _memcache_set_tasklet(self, todo, options):
    if not todo:
      raise RuntimeError('Nothing to do.')
    opname, time, namespace, deadline = options
    methodname = opname + '_multi_async'
    method = getattr(self._memcache, methodname)
    mapping = {}
    for unused_fut, (key, value) in todo:
      mapping[key] = value
    rpc = memcache.create_rpc(deadline=deadline)
    results = yield method(mapping, time=time, namespace=namespace, rpc=rpc)
    for fut, (key, unused_value) in todo:
      if results is None:
        status = memcache.MemcacheSetResponse.ERROR
      else:
        status = results.get(key)
      fut.set_result(status == memcache.MemcacheSetResponse.STORED)

  @tasklets.tasklet
  def _memcache_del_tasklet(self, todo, options):
    if not todo:
      raise RuntimeError('Nothing to do.')
    seconds, namespace, deadline = options
    keys = set()
    for unused_fut, key in todo:
      keys.add(key)
    rpc = memcache.create_rpc(deadline=deadline)
    statuses = yield self._memcache.delete_multi_async(keys, seconds=seconds,
                                                       namespace=namespace,
                                                       rpc=rpc)
    status_key_mapping = {}
    if statuses:
      for key, status in zip(keys, statuses):
        status_key_mapping[key] = status
    for fut, key in todo:
      status = status_key_mapping.get(key, memcache.DELETE_NETWORK_FAILURE)
      fut.set_result(status)

  @tasklets.tasklet
  def _memcache_off_tasklet(self, todo, options):
    if not todo:
      raise RuntimeError('Nothing to do.')
    initial_value, namespace, deadline = options
    mapping = {}
    for unused_fut, (key, delta) in todo:
      mapping[key] = delta
    rpc = memcache.create_rpc(deadline=deadline)
    results = yield self._memcache.offset_multi_async(
        mapping, initial_value=initial_value, namespace=namespace, rpc=rpc)
    for fut, (key, unused_delta) in todo:
      fut.set_result(results.get(key))

  def memcache_get(self, key, for_cas=False, namespace=None, use_cache=False,
                   deadline=None):
    """An auto-batching wrapper for `memcache.get()` or `.get_multi()`.

    Args:
      key: Key to set. This must be a string; no prefix is applied.
      for_cas: If `True`, request and store CAS ids on the Context.
      namespace: Optional namespace.
      deadline: Optional deadline for the RPC.

    Returns:
      A Future (!) whose return value is the value retrieved from
      memcache, or `None`.
    """
    if not isinstance(key, (six.text_type, six.binary_type)):
      raise TypeError('key must be a string; received %r' % key)
    if not isinstance(for_cas, bool):
      raise TypeError('for_cas must be a bool; received %r' % for_cas)
    if namespace is None:
      namespace = namespace_manager.get_namespace()
    options = (for_cas, namespace, deadline)
    batcher = self._memcache_get_batcher
    if use_cache:
      return batcher.add_once(key, options)
    else:
      return batcher.add(key, options)



  def memcache_gets(self, key, namespace=None, use_cache=False, deadline=None):
    return self.memcache_get(key, for_cas=True, namespace=namespace,
                             use_cache=use_cache, deadline=deadline)

  def memcache_set(self, key, value, time=0, namespace=None, use_cache=False,
                   deadline=None):
    if not isinstance(key, (six.text_type, six.binary_type)):
      raise TypeError('key must be a string; received %r' % key)
    if not isinstance(time, six.integer_types):
      raise TypeError('time must be a number; received %r' % time)
    if namespace is None:
      namespace = namespace_manager.get_namespace()
    options = ('set', time, namespace, deadline)
    batcher = self._memcache_set_batcher
    if use_cache:
      return batcher.add_once((key, value), options)
    else:
      return batcher.add((key, value), options)

  def memcache_add(self, key, value, time=0, namespace=None, deadline=None):
    if not isinstance(key, (six.text_type, six.binary_type)):
      raise TypeError('key must be a string; received %r' % key)
    if not isinstance(time, six.integer_types):
      raise TypeError('time must be a number; received %r' % time)
    if namespace is None:
      namespace = namespace_manager.get_namespace()
    return self._memcache_set_batcher.add((key, value),
                                          ('add', time, namespace, deadline))

  def memcache_replace(self, key, value, time=0, namespace=None, deadline=None):
    if not isinstance(key, (six.text_type, six.binary_type)):
      raise TypeError('key must be a string; received %r' % key)
    if not isinstance(time, six.integer_types):
      raise TypeError('time must be a number; received %r' % time)
    if namespace is None:
      namespace = namespace_manager.get_namespace()
    options = ('replace', time, namespace, deadline)
    return self._memcache_set_batcher.add((key, value), options)

  def memcache_cas(self, key, value, time=0, namespace=None, deadline=None):
    if not isinstance(key, (six.text_type, six.binary_type)):
      raise TypeError('key must be a string; received %r' % key)
    if not isinstance(time, six.integer_types):
      raise TypeError('time must be a number; received %r' % time)
    if namespace is None:
      namespace = namespace_manager.get_namespace()
    return self._memcache_set_batcher.add((key, value),
                                          ('cas', time, namespace, deadline))

  def memcache_delete(self, key, seconds=0, namespace=None, deadline=None):
    if not isinstance(key, (six.text_type, six.binary_type)):
      raise TypeError('key must be a string; received %r' % key)
    if not isinstance(seconds, six.integer_types):
      raise TypeError('seconds must be a number; received %r' % seconds)
    if namespace is None:
      namespace = namespace_manager.get_namespace()
    return self._memcache_del_batcher.add(key, (seconds, namespace, deadline))

  def memcache_incr(self, key, delta=1, initial_value=None, namespace=None,
                    deadline=None):
    if not isinstance(key, (six.text_type, six.binary_type)):
      raise TypeError('key must be a string; received %r' % key)
    if not isinstance(delta, six.integer_types):
      raise TypeError('delta must be a number; received %r' % delta)
    if initial_value is not None and not isinstance(initial_value,
                                                    six.integer_types):
      raise TypeError('initial_value must be a number or None; received %r' %
                      initial_value)
    if namespace is None:
      namespace = namespace_manager.get_namespace()
    return self._memcache_off_batcher.add((key, delta),
                                          (initial_value, namespace, deadline))

  def memcache_decr(self, key, delta=1, initial_value=None, namespace=None,
                    deadline=None):
    if not isinstance(key, (six.text_type, six.binary_type)):
      raise TypeError('key must be a string; received %r' % key)
    if not isinstance(delta, six.integer_types):
      raise TypeError('delta must be a number; received %r' % delta)
    if initial_value is not None and not isinstance(initial_value,
                                                    six.integer_types):
      raise TypeError('initial_value must be a number or None; received %r' %
                      initial_value)
    if namespace is None:
      namespace = namespace_manager.get_namespace()
    return self._memcache_off_batcher.add((key, -delta),
                                          (initial_value, namespace, deadline))

  @tasklets.tasklet
  def urlfetch(self, url, payload=None, method='GET', headers={},
               allow_truncated=False, follow_redirects=True,
               validate_certificate=None, deadline=None, callback=None):
    rpc = urlfetch.create_rpc(deadline=deadline, callback=callback)
    urlfetch.make_fetch_call(rpc, url,
                             payload=payload,
                             method=method,
                             headers=headers,
                             allow_truncated=allow_truncated,
                             follow_redirects=follow_redirects,
                             validate_certificate=validate_certificate)
    result = yield rpc
    raise tasklets.Return(result)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Base class for implementing API proxy stubs."""










import logging
import random
import threading

from google.appengine.api import apiproxy_rpc
from google.appengine.api import request_info
from google.appengine.runtime import apiproxy_errors

MAX_REQUEST_SIZE = 1 << 20

REQ_SIZE_EXCEEDS_LIMIT_MSG_TEMPLATE = ('The request to API call %s.%s() was too'
                                       ' large.')

logging.getLogger('google.appengine.api.stubs').setLevel(logging.INFO)


class APIProxyStub(object):
  """Base class for implementing API proxy stub classes.

  To implement an API proxy stub:

    - Extend this class.
    - Override `__init__` to pass in appropriate default service name.
    - Implement service methods as `_Dynamic_<method>(request, response)`.
  """



  _ACCEPTS_REQUEST_ID = False




  THREADSAFE = False

  def __init__(self, service_name, max_request_size=MAX_REQUEST_SIZE,
               request_data=None):
    """Constructor.

    Args:
      service_name: Service name expected for all calls.
      max_request_size: `int`. Maximum allowable size of the incoming request.
        An `apiproxy_errors.RequestTooLargeError` will be raised if the inbound
        request exceeds this size.  Default is 1 MB. Subclasses can override it.
      request_data: A `request_info.RequestInfo` instance used to look up
        state associated with the request that generated an API call.
    """
    self.__service_name = service_name
    self.__max_request_size = max_request_size
    self.request_data = request_data or request_info._local_request_info



    self._mutex = threading.RLock()
    self.__error = None
    self.__error_dict = {}

  def CreateRPC(self):
    """Creates RPC object instance.

    Returns:
      An instance of RPC.
    """
    return apiproxy_rpc.RPC(stub=self)

  def CheckRequest(self, service, call, request):
    """Check if a request meet some common restrictions.

    Args:
      service: Must be name as provided to `service_name` of constructor.
      call: A string representing the rpc to make.
      request: A protocol buffer of the type corresponding to `call`.
    """
    assert service == self.__service_name, ('Expected "%s" service name, '
                                            'was "%s"' % (self.__service_name,
                                                          service))
    if request.ByteSize() > self.__max_request_size:
      raise apiproxy_errors.RequestTooLargeError(
          REQ_SIZE_EXCEEDS_LIMIT_MSG_TEMPLATE % (service, call))
    messages = []
    assert request.IsInitialized(messages), messages

  def MakeSyncCall(self, service, call, request, response, request_id=None):
    """The main RPC entry point.

    Args:
      service: Must be name as provided to `service_name` of constructor.
      call: A string representing the rpc to make.  Must be part of
        the underlying services methods and impemented by `_Dynamic_<call>`.
      request: A protocol buffer of the type corresponding to `call`.
      response: A protocol buffer of the type corresponding to `call`.
      request_id: A unique string identifying the request associated with the
          API call.
    """
    self.CheckRequest(service, call, request)




    exception_type, frequency = self.__error_dict.get(call, (None, None))
    if exception_type and frequency:
      if random.random() <= frequency:
        raise exception_type

    if self.__error:
      if random.random() <= self.__error_rate:
        raise self.__error


    method = getattr(self, '_Dynamic_' + call)
    if self._ACCEPTS_REQUEST_ID:
      method(request, response, request_id)
    else:
      method(request, response)

  def SetError(self, error, method=None, error_rate=1):
    """Set an error condition that may be raised when calls made to stub.

    If a method is specified, the error will only apply to that call.
    The error rate is applied to the method specified or all calls if
    method is not set.

    Args:
      error: An instance of `apiproxy_errors.Error` or `None` for no error.
      method: A string representing the method that the error will affect.
      error_rate: a number from `[0, 1]` that sets the chance of the error,
        defaults to `1`.
    """
    assert error is None or isinstance(error, apiproxy_errors.Error)
    if method and error:
      self.__error_dict[method] = error, error_rate
    else:
      self.__error_rate = error_rate
      self.__error = error


def Synchronized(method):
  """Decorator to acquire a mutex around an `APIProxyStub` method.

  Args:
    method: An unbound method of `APIProxyStub` or a subclass.

  Returns:
    The `method`, altered such it acquires `self._mutex` throughout its
    execution.
  """

  def WrappedMethod(self, *args, **kwargs):
    with self._mutex:
      return method(self, *args, **kwargs)

  return WrappedMethod



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
















"""The Key class, and associated utilities.

A Key encapsulates the following pieces of information, which together
uniquely designate a (possible) entity in the App Engine datastore:

- an application id (a string)
- a namespace (a string)
- a list of one or more (kind, id) pairs where kind is a string and id
  is either a string or an integer.

The application id must always be part of the key, but since most
applications can only access their own entities, it defaults to the
current application id and you rarely need to worry about it.  It must
not be empty.

The namespace designates a top-level partition of the key space for a
particular application.  If you've never heard of namespaces, you can
safely ignore this feature.

Most of the action is in the (kind, id) pairs.  A key must have at
least one (kind, id) pair.  The last (kind, id) pair gives the kind
and the id of the entity that the key refers to, the others merely
specify a 'parent key'.

The kind is a string giving the name of the model class used to
represent the entity.  (In more traditional databases this would be
the table name.)  A model class is a Python class derived from
ndb.Model; see the documentation for ndb/model.py.  Only the class
name itself is used as the kind.  This means all your model classes
must be uniquely named within one application.  You can override this
on a per-class basis.

The id is either a string or an integer.  When the id is a string, the
application is in control of how it assigns ids: For example, if you
could use an email address as the id for Account entities.

To use integer ids, you must let the datastore choose a unique id for
an entity when it is first inserted into the datastore.  You can set
the id to None to represent the key for an entity that hasn't yet been
inserted into the datastore.  The final key (including the assigned
id) will be returned after the entity is successfully inserted into
the datastore.

A key for which the id of the last (kind, id) pair is set to None is
called an incomplete key.  Such keys can only be used to insert
entities into the datastore.

A key with exactly one (kind, id) pair is called a top level key or a
root key.  Top level keys are also used as entity groups, which play a
role in transaction management.

If there is more than one (kind, id) pair, all but the last pair
represent the 'ancestor path', also known as the key of the 'parent
entity'.

Other constraints:

- Kinds and string ids must not be empty and must be at most 500 bytes
  long (after UTF-8 encoding, if given as Python unicode objects).
  NOTE: This is defined as a module level constant _MAX_KEYPART_BYTES.

- Integer ids must be at least 1 and less than 2**63.

For more info about namespaces, see
http://code.google.com/appengine/docs/python/multitenancy/overview.html.
The namespace defaults to the 'default namespace' selected by the
namespace manager.  To explicitly select the empty namespace pass
namespace=''.
"""

import base64
import os

from google.appengine.api import cmp_compat
from google.appengine.api import datastore_errors
from google.appengine.api import datastore_types
from google.appengine.api import full_app_id
from google.appengine.api import namespace_manager
from google.appengine.ext.ndb import utils
import six
from six.moves import range

from google.appengine.datastore import entity_bytes_pb2 as entity_pb2

__all__ = ['Key']

_MAX_LONG = 2**63
_MAX_KEYPART_BYTES = 500




@cmp_compat.total_ordering_from_cmp
class Key(object):
  """An immutable datastore key.

  For flexibility and convenience, multiple constructor signatures are
  supported.

  The primary way to construct a key is using positional arguments:
  - Key(kind1, id1, kind2, id2, ...).

  This is shorthand for either of the following two longer forms:
  - Key(pairs=[(kind1, id1), (kind2, id2), ...])
  - Key(flat=[kind1, id1, kind2, id2, ...])

  Either of the above constructor forms can additionally pass in another
  key using parent=<key>.  The (kind, id) pairs of the parent key are
  inserted before the (kind, id) pairs passed explicitly.

  You can also construct a Key from a 'url-safe' encoded string:
  - Key(urlsafe=<string>)

  For esoteric purposes the following constructors exist:
  - Key(reference=<reference>) -- passing in a low-level Reference object
  - Key(serialized=<string>) -- passing in a serialized low-level Reference
  - Key(<dict>) -- for unpickling, the same as Key(**<dict>)

  The 'url-safe' string is really a websafe-base64-encoded serialized
  Reference, but it's best to think of it as just an opaque unique
  string.

  Additional constructor keyword arguments:
  - app=<string> -- specify the application id
  - namespace=<string> -- specify the namespace

  If a Reference is passed (using one of reference, serialized or
  urlsafe), the args and namespace keywords must match what is already
  present in the Reference (after decoding if necessary).  The parent
  keyword cannot be combined with a Reference in any form.


  Keys are immutable, which means that a Key object cannot be modified
  once it has been created.  This is enforced by the implementation as
  well as Python allows.

  For access to the contents of a key, the following methods and
  operations are supported:

  - repr(key), str(key) -- return a string representation resembling
    the shortest constructor form, omitting the app and namespace
    unless they differ from the default value.

  - key1 == key2, key1 != key2 -- comparison for equality between Keys.

  - hash(key) -- a hash value sufficient for storing Keys in a dict.

  - key.pairs() -- a tuple of (kind, id) pairs.

  - key.flat() -- a tuple of flattened kind and id values, i.e.
    (kind1, id1, kind2, id2, ...).

  - key.app() -- the application id.

  - key.id() -- the string or integer id in the last (kind, id) pair,
    or None if the key is incomplete.

  - key.string_id() -- the string id in the last (kind, id) pair,
    or None if the key has an integer id or is incomplete.

  - key.integer_id() -- the integer id in the last (kind, id) pair,
    or None if the key has a string id or is incomplete.

  - key.namespace() -- the namespace.

  - key.kind() -- a shortcut for key.pairs()[-1][0].

  - key.parent() -- a Key constructed from all but the last (kind, id)
    pairs.

  - key.urlsafe() -- a websafe-base64-encoded serialized Reference.

  - key.serialized() -- a serialized Reference.

  - key.reference() -- a Reference object.  The caller promises not to
    mutate it.

  Keys also support interaction with the datastore; these methods are
  the only ones that engage in any kind of I/O activity.  For Future
  objects, see the document for ndb/tasklets.py.

  - key.get() -- return the entity for the Key.

  - key.get_async() -- return a Future whose eventual result is
    the entity for the Key.

  - key.delete() -- delete the entity for the Key.

  - key.delete_async() -- asynchronously delete the entity for the Key.

  Keys may be pickled.

  Subclassing Key is best avoided; it would be hard to get right.
  """

  __slots__ = ['__reference', '__pairs', '__app', '__namespace']

  def __new__(cls, *_args, **kwargs):
    """Constructor.  See the class docstring for arguments."""
    if _args:
      if len(_args) == 1 and isinstance(_args[0], dict):
        if kwargs:
          raise TypeError('Key() takes no keyword arguments when a dict is the '
                          'the first and only non-keyword argument (for '
                          'unpickling).')
        kwargs = _args[0]
      else:
        if 'flat' in kwargs:
          raise TypeError('Key() with positional arguments '
                          'cannot accept flat as a keyword argument.')
        kwargs['flat'] = _args


    if int(os.environ.get('NDB_PY2_UNPICKLE_COMPAT', '0')):
      kwargs = {six.ensure_str(k): v for (k, v) in kwargs.items()}

    self = super(Key, cls).__new__(cls)



    if 'reference' in kwargs or 'serialized' in kwargs or 'urlsafe' in kwargs:
      (self.__reference,
       self.__pairs,
       self.__app,
       self.__namespace) = self._parse_from_ref(cls, **kwargs)
    elif 'pairs' in kwargs or 'flat' in kwargs:
      self.__reference = None
      (self.__pairs,
       self.__app,
       self.__namespace) = self._parse_from_args(**kwargs)
    else:
      raise TypeError('Key() cannot create a Key instance without arguments.')
    return self

  @staticmethod
  def _parse_from_args(pairs=None, flat=None, app=None, namespace=None,
                       parent=None):
    if flat:
      if pairs is not None:
        raise TypeError('Key() cannot accept both flat and pairs arguments.')
      if len(flat) % 2:
        raise ValueError('Key() must have an even number of positional '
                         'arguments.')
      pairs = [(flat[i], flat[i + 1]) for i in range(0, len(flat), 2)]
    else:
      pairs = list(pairs)
    if not pairs:
      raise TypeError('Key must consist of at least one pair.')
    for i, (kind, id) in enumerate(pairs):
      if isinstance(id, six.text_type):
        id = six.ensure_binary(id)
      elif id is None:
        if i + 1 < len(pairs):
          raise datastore_errors.BadArgumentError(
              'Incomplete Key entry must be last')
      else:
        if not isinstance(id, six.integer_types + (six.binary_type,)):
          raise TypeError('Key id must be a string or a number; received %r' %
                          id)
      if isinstance(kind, type):
        kind = kind._get_kind()
      if isinstance(kind, six.text_type):
        kind = six.ensure_binary(kind)
      if not isinstance(kind, six.binary_type):
        raise TypeError(
            'Key kind must be a string or Model class; received %r' % kind)
      if not id:
        id = None
      pairs[i] = (kind, id)
    if parent is not None:
      if not isinstance(parent, Key):
        raise datastore_errors.BadValueError(
            'Expected Key instance, got %r' % parent)
      if not parent.id():
        raise datastore_errors.BadArgumentError(
            'Parent cannot have incomplete key')
      pairs[:0] = parent.__pairs
      if app:
        if app != parent.app():
          raise ValueError('Cannot specify a different app %r '
                           'than the parent app %r' %
                           (app, parent.app()))
      else:
        app = parent.__app
      if namespace is not None:
        if namespace != parent.namespace():
          raise ValueError('Cannot specify a different namespace %r '
                           'than the parent namespace %r' %
                           (namespace, parent.namespace()))
      else:
        namespace = parent.__namespace
    if not app:
      app = _DefaultAppId()
    if namespace is None:
      namespace = namespace_manager.get_namespace()



    app = six.ensure_binary(app)
    namespace = six.ensure_binary(namespace)

    return tuple(pairs), app, namespace

  @staticmethod
  def _parse_from_ref(cls, pairs=None, flat=None,
                      reference=None, serialized=None, urlsafe=None,
                      app=None, namespace=None, parent=None):
    """Construct a Reference; the signature is the same as for Key."""
    if cls is not Key:
      raise TypeError('Cannot construct Key reference on non-Key class; '
                      'received %r' % cls)
    if (bool(pairs) + bool(flat) + bool(reference) + bool(serialized) +
        bool(urlsafe) + bool(parent)) != 1:
      raise TypeError('Cannot construct Key reference from incompatible '
                      'keyword arguments.')
    if urlsafe:
      serialized = _DecodeUrlSafe(urlsafe)
    if serialized:
      reference = _ReferenceFromSerialized(serialized)
    if reference:
      reference = _ReferenceFromReference(reference)
    pairs = []
    elem = None
    path = reference.path
    for elem in path.element:
      kind = six.ensure_binary(elem.type)
      if elem.HasField('id'):
        id_or_name = elem.id
      else:
        id_or_name = six.ensure_binary(elem.name)
      if not id_or_name:
        id_or_name = None
      tup = (kind, id_or_name)
      pairs.append(tup)
    if elem is None:
      raise RuntimeError('Key reference has no path or elements (%r, %r, %r).'
                         % (urlsafe, serialized, str(reference)))




    ref_app = reference.app
    if app is not None:
      if app != ref_app:
        raise RuntimeError('Key reference constructed uses a different app %r '
                           'than the one specified %r' %
                           (ref_app, app))
    ref_namespace = reference.name_space
    if namespace is not None:
      if namespace != ref_namespace:
        raise RuntimeError('Key reference constructed uses a different '
                           'namespace %r than the one specified %r' %
                           (ref_namespace, namespace))



    ref_app = six.ensure_binary(ref_app)
    ref_namespace = six.ensure_binary(ref_namespace)

    return (reference, tuple(pairs), ref_app, ref_namespace)

  def _bytes2str(self, val):
    if isinstance(val, bytes):
      val = six.ensure_str(val)
    return val

  def __repr__(self):
    """String representation, used by str() and repr().

    We produce a short string that conveys all relevant information,
    suppressing app and namespace when they are equal to the default.
    """

    args = []
    for item in self.flat():
      if not item:
        args.append('None')
      elif isinstance(item, six.string_types):
        args.append(repr(item))
      else:
        args.append(str(item))
    if self.app() != _DefaultAppId():
      args.append('app=%r' % self.app())
    if self.namespace() != namespace_manager.get_namespace():
      args.append('namespace=%r' % self.namespace())
    return 'Key(%s)' % ', '.join(args)

  __str__ = __repr__

  def __hash__(self):
    """Hash value, for use in dict lookups."""




    return hash(tuple(self.pairs()))

  def __eq__(self, other):
    """Equality comparison operation."""


    if not isinstance(other, Key):
      return NotImplemented
    return (self.__pairs == other.__pairs and
            self.__app == other.__app and
            self.__namespace == other.__namespace)

  def __ne__(self, other):
    """The opposite of __eq__."""
    if not isinstance(other, Key):
      return NotImplemented
    return not self.__eq__(other)

  def __cmp__(self, other):

    if not isinstance(other, Key):
      return NotImplemented


    tuple_1 = (
        self.__app,
        self.__namespace,
        _ConvertPairsForComparison(self.__pairs))
    tuple_2 = (
        other.__app,
        other.__namespace,
        _ConvertPairsForComparison(other.__pairs))

    return cmp_compat.cmp(tuple_1, tuple_2)

  def __getstate__(self):
    """Private API used for pickling."""


    return ({'pairs': self.__pairs,
             'app': self.__app,
             'namespace': self.__namespace},)

  def __setstate__(self, state):
    """Private API used for pickling."""
    if len(state) != 1:
      raise TypeError('Invalid state length, expected 1; received %i' %
                      len(state))
    kwargs = state[0]
    if not isinstance(kwargs, dict):
      raise TypeError('Key accepts a dict of keyword arguments as state; '
                      'received %r' % kwargs)
    kwargs = {six.ensure_text(k): v for k, v in kwargs.items()}
    self.__reference = None
    (self.__pairs, self.__app, self.__namespace) = self._parse_from_args(
        pairs=tuple(kwargs['pairs']),
        app=kwargs['app'],
        namespace=kwargs['namespace'])

  def __getnewargs__(self):
    """Private API used for pickling."""
    return ({'pairs': self.__pairs,
             'app': self.__app,
             'namespace': self.__namespace},)

  def parent(self):
    """Return a Key constructed from all but the last (kind, id) pairs.

    If there is only one (kind, id) pair, return None.
    """
    pairs = self.__pairs
    if len(pairs) <= 1:
      return None
    return Key(pairs=pairs[:-1], app=self.__app, namespace=self.__namespace)

  def root(self):
    """Return the root key.  This is either self or the highest parent."""
    pairs = self.__pairs
    if len(pairs) <= 1:
      return self
    return Key(pairs=pairs[:1], app=self.__app, namespace=self.__namespace)

  def namespace(self):
    """Return the namespace."""
    return self._bytes2str(self.__namespace)

  def app(self):
    """Return the application id."""
    return self._bytes2str(self.__app)

  def id(self):
    """Return the string or integer id in the last (kind, id) pair, if any.

    Returns:
      A string or integer id, or None if the key is incomplete.
    """
    return self._bytes2str(self.__pairs[-1][1])

  def string_id(self):
    """Return the string id in the last (kind, id) pair, if any.

    Returns:
      A string id, or None if the key has an integer id or is incomplete.
    """
    id = self.id()
    if not isinstance(id, (str, bytes)):
      id = None
    return id

  def integer_id(self):
    """Return the integer id in the last (kind, id) pair, if any.

    Returns:
      An integer id, or None if the key has a string id or is incomplete.
    """
    id = self.id()
    if not isinstance(id, six.integer_types):
      id = None
    return id

  def pairs(self):
    """Return a tuple of (kind, id) pairs."""
    return tuple((self._bytes2str(p[0]), self._bytes2str(p[1]))
                 for p in self.__pairs)

  def flat(self):
    """Return a tuple of alternating kind and id values."""
    flat = []
    for kind, ID in self.pairs():
      flat.append(kind)
      flat.append(ID)
    return tuple(flat)

  def kind(self):
    """Return the kind of the entity referenced.

    This is the kind from the last (kind, id) pair.
    """
    return self._bytes2str(self.__pairs[-1][0])

  def reference(self):
    """Return the Reference object for this Key.

    This is a entity_pb2.Reference instance -- a protocol buffer class
    used by the lower-level API to the datastore.

    NOTE: The caller should not mutate the return value.
    """
    if self.__reference is None:
      self.__reference = _ConstructReference(self.__class__,
                                             pairs=self.__pairs,
                                             app=self.__app,
                                             namespace=self.__namespace)
    return self.__reference

  def serialized(self):
    """Return a serialized Reference object for this Key."""
    return self.reference().SerializeToString()

  def urlsafe(self):
    """Return a url-safe string encoding this Key's Reference.

    This string is compatible with other APIs and languages and with
    the strings used to represent Keys in GQL and in the App Engine
    Admin Console.
    """

    urlsafe = base64.b64encode(self.reference().SerializeToString())
    return urlsafe.rstrip(b'=').replace(b'+', b'-').replace(b'/', b'_')




  def get(self, **ctx_options):
    """Synchronously get the entity for this Key.

    Return None if there is no such entity.
    """
    return self.get_async(**ctx_options).get_result()

  def get_async(self, **ctx_options):
    """Return a Future whose result is the entity for this Key.

    If no such entity exists, a Future is still returned, and the
    Future's eventual return result be None.
    """
    from google.appengine.ext.ndb import model, tasklets
    ctx = tasklets.get_context()
    cls = model.Model._kind_map.get(self.kind())
    if cls:
      cls._pre_get_hook(self)
    fut = ctx.get(self, **ctx_options)
    if cls:
      post_hook = cls._post_get_hook
      if not cls._is_default_hook(model.Model._default_post_get_hook,
                                  post_hook):
        fut.add_immediate_callback(post_hook, self, fut)
    return fut

  def delete(self, **ctx_options):
    """Synchronously delete the entity for this Key.

    This is a no-op if no such entity exists.
    """
    return self.delete_async(**ctx_options).get_result()

  def delete_async(self, **ctx_options):
    """Schedule deletion of the entity for this Key.

    This returns a Future, whose result becomes available once the
    deletion is complete.  If no such entity exists, a Future is still
    returned.  In all cases the Future's result is None (i.e. there is
    no way to tell whether the entity existed or not).
    """
    from google.appengine.ext.ndb import tasklets, model
    ctx = tasklets.get_context()
    cls = model.Model._kind_map.get(self.kind())
    if cls:
      cls._pre_delete_hook(self)
    fut = ctx.delete(self, **ctx_options)
    if cls:
      post_hook = cls._post_delete_hook
      if not cls._is_default_hook(model.Model._default_post_delete_hook,
                                  post_hook):
        fut.add_immediate_callback(post_hook, self, fut)
    return fut

  @classmethod
  def from_old_key(cls, old_key):
    return cls(urlsafe=str(old_key))

  def to_old_key(self):
    return datastore_types.Key(encoded=self.urlsafe())





@utils.positional(1)
def _ConstructReference(cls, pairs=None, flat=None,
                        reference=None, serialized=None, urlsafe=None,
                        app=None, namespace=None, parent=None):
  """Construct a Reference; the signature is the same as for Key."""
  if cls is not Key:
    raise TypeError('Cannot construct Key reference on non-Key class; '
                    'received %r' % cls)
  if (bool(pairs) + bool(flat) + bool(reference) + bool(serialized) +
      bool(urlsafe)) != 1:
    raise TypeError('Cannot construct Key reference from incompatible keyword '
                    'arguments.')
  if flat or pairs:
    if flat:
      if len(flat) % 2:
        raise TypeError('_ConstructReference() must have an even number of '
                        'positional arguments.')
      pairs = [(flat[i], flat[i + 1]) for i in range(0, len(flat), 2)]
    elif parent is not None:
      pairs = list(pairs)
    if not pairs:
      raise TypeError('Key references must consist of at least one pair.')
    if parent is not None:
      if not isinstance(parent, Key):
        raise datastore_errors.BadValueError(
            'Expected Key instance, got %r' % parent)
      pairs[:0] = parent.pairs()
      if app:
        if app != parent.app():
          raise ValueError('Cannot specify a different app %r '
                           'than the parent app %r' %
                           (app, parent.app()))
      else:
        app = parent.app()
      if namespace is not None:
        if namespace != parent.namespace():
          raise ValueError('Cannot specify a different namespace %r '
                           'than the parent namespace %r' %
                           (namespace, parent.namespace()))
      else:
        namespace = parent.namespace()
    reference = _ReferenceFromPairs(pairs, app=app, namespace=namespace)
  else:
    if parent is not None:
      raise TypeError('Key reference cannot be constructed when the parent '
                      'argument is combined with either reference, serialized '
                      'or urlsafe arguments.')
    if urlsafe:
      serialized = _DecodeUrlSafe(urlsafe)
    if serialized:
      reference = _ReferenceFromSerialized(serialized)
    if not reference.path().element_size():
      raise RuntimeError('Key reference has no path or elements (%r, %r, %r).'
                         % (urlsafe, serialized, str(reference)))

    if not serialized:
      reference = _ReferenceFromReference(reference)



    if app is not None:
      ref_app = reference.app()
      if app != ref_app:
        raise RuntimeError('Key reference constructed uses a different app %r '
                           'than the one specified %r' %
                           (ref_app, app))
    if namespace is not None:
      ref_namespace = reference.name_space()
      if namespace != ref_namespace:
        raise RuntimeError('Key reference constructed uses a different '
                           'namespace %r than the one specified %r' %
                           (ref_namespace, namespace))
  return reference


def _ReferenceFromPairs(pairs, reference=None, app=None, namespace=None):
  """Construct a Reference from a list of pairs.

  If a Reference is passed in as the second argument, it is modified
  in place.  The app and namespace are set from the corresponding
  keyword arguments, with the customary defaults.
  """
  if reference is None:
    reference = entity_pb2.Reference()
  reference.path.SetInParent()
  path = reference.path
  last = False
  for kind, idorname in pairs:
    if last:
      raise datastore_errors.BadArgumentError(
          'Incomplete Key entry must be last')
    t = type(kind)
    if t is six.binary_type:
      pass
    elif t is six.text_type:
      kind = six.ensure_binary(kind)
    else:
      if issubclass(t, type):

        from google.appengine.ext.ndb.model import Model
        modelclass = kind
        if not issubclass(modelclass, Model):
          raise TypeError('Key kind must be either a string or subclass of '
                          'Model; received %r' % modelclass)
        kind = modelclass._get_kind()
        t = type(kind)
      if t is six.binary_type:
        pass
      elif t is six.text_type:
        kind = six.ensure_binary(kind)
      elif issubclass(t, six.binary_type):
        pass
      elif issubclass(t, six.text_type):
        kind = six.ensure_binary(kind)
      else:
        raise TypeError('Key kind must be either a string or subclass of Model;'
                        ' received %r' % kind)

    if not (1 <= len(kind) <= _MAX_KEYPART_BYTES):
      raise ValueError('Key kind string must be a non-empty string up to %i'
                       'bytes; received %s' %
                       (_MAX_KEYPART_BYTES, kind))
    elem = path.element.add()
    elem.type = kind
    t = type(idorname)
    if t is int or t is int:

      if not (1 <= idorname < _MAX_LONG):
        raise ValueError('Key id number is too long; received %i' % idorname)
      elem.id = idorname
    elif t is str:

      if not (1 <= len(idorname) <= _MAX_KEYPART_BYTES):
        raise ValueError('Key name strings must be non-empty strings up to %i '
                         'bytes; received %s' %
                         (_MAX_KEYPART_BYTES, idorname))
      elem.name = idorname
    elif t is six.text_type:
      idorname = idorname.encode('utf8')

      if not (1 <= len(idorname) <= _MAX_KEYPART_BYTES):
        raise ValueError('Key name unicode strings must be non-empty strings up'
                         ' to %i bytes; received %s' %
                         (_MAX_KEYPART_BYTES, idorname))
      elem.name = idorname
    elif idorname is None:
      last = True
    elif issubclass(t, six.integer_types):

      if not (1 <= idorname < _MAX_LONG):
        raise ValueError('Key id number is too long; received %i' % idorname)
      elem.id = idorname
    elif issubclass(t, six.string_types) or t is six.binary_type:
      if issubclass(t, six.text_type):
        idorname = idorname.encode('utf8')

      if not (1 <= len(idorname) <= _MAX_KEYPART_BYTES):
        raise ValueError('Key name strings must be non-empty strings up to %i '
                         'bytes; received %s' % (_MAX_KEYPART_BYTES, idorname))
      elem.name = idorname
    else:
      raise TypeError('id must be either a numeric id or a string name; '
                      'received %r' % idorname)

  if not app:
    app = _DefaultAppId()

  reference.app = app

  if namespace is None:
    namespace = namespace_manager.get_namespace()

  if namespace:
    reference.name_space = namespace
  return reference


def _ReferenceFromReference(reference):
  """Copy a Reference."""
  new_reference = entity_pb2.Reference()
  new_reference.CopyFrom(reference)
  return new_reference


def _ReferenceFromSerialized(serialized):
  """Construct a Reference from a serialized Reference."""
  if not isinstance(serialized, (six.text_type, six.binary_type)):
    raise TypeError('serialized must be a string; received %r' % serialized)
  elif isinstance(serialized, six.text_type):
    serialized = serialized.encode('utf8')
  return entity_pb2.Reference.FromString(serialized)


def _DecodeUrlSafe(urlsafe):
  """Decode a url-safe base64-encoded string.

  This returns the decoded string.
  """
  if not isinstance(urlsafe, (six.text_type, six.binary_type)):
    raise TypeError('urlsafe must be a string; received %r' % urlsafe)
  if isinstance(urlsafe, six.text_type):
    urlsafe = urlsafe.encode('utf8')
  mod = len(urlsafe) % 4
  if mod:
    urlsafe += b'=' * (4 - mod)

  return base64.b64decode(urlsafe.replace(b'-', b'+').replace(b'_', b'/'))


def _DefaultAppId():
  """Return the default application id.

  This is taken from environment variables.
  """
  return full_app_id.get()


def _ConvertPairsForComparison(pairs):
  """Converts Key pairs to a format that enables consistent comparison.

  Here's a nasty one. Due to the tighter constraints on what can be compared
  in PY3, we can't compare Keys tuples directly, as they can contain a mix of
  strings and integers. The PY2 behavior in such cases seems to be that in
  cases of differing types, comparisons are determined via the type name
  itself.

  Now the nasty part. While this could be addressed by something along the
  lines of...

    cmp(str(type(x)), str(type(y)))

  ...in PY2, this results in 'int' < 'str'. But in PY3, this results in
  'bytes' < 'int', which produces a different sorted ordering of said Keys. In
  the interest of minimizing breakage in existing code, I implemented this
  method to maintain the PY2 sort ordering for Keys. Comparing mixed items in
  PY2 now results in 'int' < 'unicode', and PY3 results in 'int' < 'str'.

  Args:
    pairs: A tuple of tuples, corresponding to the (kind, id) pairs of a Key.

  Returns:
    A tuple of tuples, converted for consistent comparison.
  """
  conv = lambda x: six.ensure_text(x) if isinstance(x, six.binary_type) else x
  return tuple((conv(i), conv(j)) for i, j in pairs)
